# Copyright (C) 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# Author: Aaron Bentley <aaron.bentley@utoronto.ca>
from breezy.tests import TestCase
from breezy.textmerge import Merge2


class TestMerge2(TestCase):
    def test_agreed(self):
        lines = "a\nb\nc\nd\ne\nf\n".splitlines(True)
        mlines = list(Merge2(lines, lines).merge_lines()[0])
        self.assertEqualDiff(mlines, lines)

    def test_conflict(self):
        lines_a = "a\nb\nc\nd\ne\nf\ng\nh\n".splitlines(True)
        lines_b = "z\nb\nx\nd\ne\ne\nf\ng\ny\n".splitlines(True)
        expected = "<\na\n=\nz\n>\nb\n<\nc\n=\nx\n>\nd\ne\n<\n=\ne\n>\nf\n"\
                   "g\n<\nh\n=\ny\n>\n"
        m2 = Merge2(lines_a, lines_b, '<\n', '>\n', '=\n')
        mlines = m2.merge_lines()[0]
        self.assertEqualDiff(''.join(mlines), expected)
        mlines = m2.merge_lines(reprocess=True)[0]
        self.assertEqualDiff(''.join(mlines), expected)

    def test_reprocess(self):
        struct = [('a', 'b'), ('c',), ('def', 'geh'), ('i',)]
        expect = [('a', 'b'), ('c',), ('d', 'g'), ('e',), ('f', 'h'), ('i',)]
        result = Merge2.reprocess_struct(struct)
        self.assertEqual(list(result), expect)



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Commands for generating snapshot information about a brz tree."""

from .lazy_import import lazy_import

lazy_import(globals(), """
from breezy import (
    branch,
    version_info_formats,
    workingtree,
    )
from breezy.i18n import gettext
""")

from . import errors
from .commands import Command
from .option import Option, RegistryOption


def _parse_version_info_format(format):
    """Convert a string passed by the user into a VersionInfoFormat.

    This looks in the version info format registry, and if the format
    cannot be found, generates a useful error exception.
    """
    try:
        return version_info_formats.get_builder(format)
    except KeyError:
        formats = version_info_formats.get_builder_formats()
        raise errors.CommandError(
            gettext('No known version info format {0}.'
                    ' Supported types are: {1}').format(format, formats))


class cmd_version_info(Command):
    __doc__ = """Show version information about this tree.

    You can use this command to add information about version into
    source code of an application. The output can be in one of the
    supported formats or in a custom format based on a template.

    For example::

      brz version-info --custom \\
        --template="#define VERSION_INFO \\"Project 1.2.3 (r{revno})\\"\\n"

    will produce a C header file with formatted string containing the
    current revision number. Other supported variables in templates are:

      * {date} - date of the last revision
      * {build_date} - current date
      * {revno} - revision number
      * {revision_id} - revision id
      * {branch_nick} - branch nickname
      * {clean} - 0 if the source tree contains uncommitted changes,
                  otherwise 1
    """

    takes_options = [RegistryOption('format',
                                    'Select the output format.',
                                    value_switches=True,
                                    lazy_registry=('breezy.version_info_formats',
                                                   'format_registry')),
                     Option('all', help='Include all possible information.'),
                     Option('check-clean', help='Check if tree is clean.'),
                     Option('include-history',
                            help='Include the revision-history.'),
                     Option('include-file-revisions',
                            help='Include the last revision for each file.'),
                     Option('template', type=str,
                            help='Template for the output.'),
                     'revision',
                     ]
    takes_args = ['location?']

    encoding_type = 'replace'

    def run(self, location=None, format=None,
            all=False, check_clean=False, include_history=False,
            include_file_revisions=False, template=None,
            revision=None):

        if revision and len(revision) > 1:
            raise errors.CommandError(
                gettext('brz version-info --revision takes exactly'
                        ' one revision specifier'))

        if location is None:
            location = '.'

        if format is None:
            format = version_info_formats.format_registry.get()

        try:
            wt = workingtree.WorkingTree.open_containing(location)[0]
        except errors.NoWorkingTree:
            b = branch.Branch.open(location)
            wt = None
        else:
            b = wt.branch

        if all:
            include_history = True
            check_clean = True
            include_file_revisions = True
        if template:
            include_history = True
            include_file_revisions = True
            if '{clean}' in template:
                check_clean = True

        if revision is not None:
            revision_id = revision[0].as_revision_id(b)
        else:
            revision_id = None

        builder = format(b, working_tree=wt,
                         check_for_clean=check_clean,
                         include_revision_history=include_history,
                         include_file_revisions=include_file_revisions,
                         template=template, revision_id=revision_id)
        builder.generate(self.outf)



# Copyright (C) 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the InterTree.compare() function."""

import os
import shutil

from breezy import errors, mutabletree, tests
from breezy.bzr.inventorytree import InventoryTreeChange
from breezy.osutils import supports_symlinks
from breezy.tests import features
from breezy.tests.matchers import MatchesTreeChanges
from breezy.tests.per_intertree import TestCaseWithTwoTrees
from breezy.tree import TreeChange

# TODO: test the include_root option.
# TODO: test that renaming a directory x->y does not emit a rename for the
#       child x/a->y/a.
# TODO: test that renaming a directory x-> does not emit a rename for the child
#        x/a -> y/a when a supplied_files argument gives either 'x/' or 'y/a'
#        -> that is, when the renamed parent is not processed by the function.
# TODO: test items are only emitted once when a specific_files list names a dir
#       whose parent is now a child.
# TODO: test comparisons between trees with different root ids. mbp 20070301
#
# TODO: More comparisons between trees with subtrees in different states.
#
# TODO: Many tests start out by setting the tree roots ids the same, maybe
#       that should just be the default for these tests, by changing
#       make_branch_and_tree.  mbp 20070307


def _change_key(change):
    """Return a valid key for sorting Tree.iter_changes entries."""
    return (change.file_id or b'', (change.path[0] or '', change.path[1] or ''),
            change.versioned, change.parent_id, change.name, change.kind,
            change.executable)


class TestCompare(TestCaseWithTwoTrees):

    def _make_abc_tree(self, tree):
        """setup an abc content tree."""
        files = ['a', 'b/', 'b/c']
        self.build_tree(files, line_endings='binary',
                        transport=tree.controldir.root_transport)
        tree.set_root_id(b'root-id')
        tree.add(files, ids=[b'a-id', b'b-id', b'c-id'])

    def get_tree_no_parents_abc_content(self, tree, converter=None):
        """return a test tree with a, b/, b/c contents."""
        self._make_abc_tree(tree)
        return self._convert_tree(tree, converter)

    def test_compare_empty_trees(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_no_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_empty_to_abc_content(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual(
            [('a', 'file'), ('b', 'directory'), ('b/c', 'file')],
            [(c.path[1], c.kind[1]) for c in d.added])
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_dangling(self):
        # This test depends on the ability for some trees to have a difference
        # between a 'versioned present' and 'versioned not present' (aka
        # dangling) file. In this test there are two trees each with a separate
        # dangling file, and the dangling files should be considered absent for
        # the test.
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['2/a'])
        tree2.add('a')
        os.unlink('2/a')
        self.build_tree(['1/b'])
        tree1.add('b')
        os.unlink('1/b')
        # the conversion to test trees here will leave the trees intact for the
        # default intertree, but may perform a commit for other tree types,
        # which may reduce the validity of the test. XXX: Think about how to
        # address this.
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_abc_content_to_empty(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_no_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([('a', 'file'),
                          ('b', 'directory'),
                          ('b/c', 'file'),
                          ], [(c.path[0], c.kind[0]) for c in d.removed])
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_content_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_2(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual(
            [('a', 'file', True, False)],
            [(c.path[1], c.kind[1], c.changed_content, c.meta_modified())
             for c in d.modified])
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_meta_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_3(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual(
            [('b/c', 'file', False, True)],
            [(c.path[1], c.kind[1], c.changed_content, c.meta_modified())
             for c in d.modified])
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_file_rename(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_4(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual(
            [('a', 'd', 'file', False, False)],
            [(c.path[0], c.path[1], c.kind[1], c.changed_content, c.meta_modified())
             for c in d.renamed])
        self.assertEqual([], d.unchanged)

    def test_file_rename_and_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_5(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual(
            [('a', 'd', 'file', True, False)],
            [(c.path[0], c.path[1], c.kind[1], c.changed_content, c.meta_modified())
             for c in d.renamed])
        self.assertEqual([], d.unchanged)

    def test_file_rename_and_meta_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_6(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual(
            [('b/c', 'e', 'file', False, True)],
            [(c.path[0], c.path[1], c.kind[1], c.changed_content, c.meta_modified())
             for c in d.renamed])
        self.assertEqual([], d.unchanged)

    def test_empty_to_abc_content_a_only(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare(specific_files=['a'])
        self.assertEqual(
            [('a', 'file')],
            [(c.path[1], c.kind[1]) for c in d.added])
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_empty_to_abc_content_a_and_c_only(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare(
            specific_files=['a', 'b/c'])
        self.assertEqual(
            [('a', 'file'),
             ('b', 'directory'),
             ('b/c', 'file')],
            [(c.path[1], c.kind[1]) for c in d.added])
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_empty_to_abc_content_c_only(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare(
            specific_files=['b/c'])
        self.assertEqual(
            [('b', 'directory'), ('b/c', 'file')],
            [(c.path[1], c.kind[1]) for c in d.added])
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_empty_to_abc_content_b_only(self):
        """Restricting to a dir matches the children of the dir."""
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare(specific_files=['b'])
        self.assertEqual(
            [('b', 'directory'), ('b/c', 'file')],
            [(c.path[1], c.kind[1]) for c in d.added])
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_unchanged_with_renames_and_modifications(self):
        """want_unchanged should generate a list of unchanged entries."""
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_5(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare(want_unchanged=True)
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual(
            [('a', 'd', 'file', True, False)],
            [(c.path[0], c.path[1], c.kind[1], c.changed_content, c.meta_modified()) for c in d.renamed])
        self.assertEqual(
            [('b', 'directory'), ('b/c', 'file')],
            [(c.path[0], c.kind[0]) for c in d.unchanged])

    def test_extra_trees_finds_ids(self):
        """Ask for a delta between two trees with a path present in a third."""
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_3(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare(specific_files=['b'])
        # the type of tree-3 does not matter - it is used as a lookup, not
        # a dispatch. XXX: For dirstate it does speak to the optimisability of
        # the lookup, in merged trees it can be fast-pathed. We probably want
        # two tests: one as is, and one with it as a pending merge.
        tree3 = self.make_branch_and_tree('3')
        tree3 = self.get_tree_no_parents_abc_content_6(tree3)
        tree3.lock_read()
        self.addCleanup(tree3.unlock)
        # tree 3 has 'e' which is b'c-id'. Tree 1 has c-id at b/c, and Tree 2
        # has c-id at b/c with its exec flag toggled.
        # without extra_trees, we should get no modifications from this
        # so do one, to be sure the test is valid.
        d = self.intertree_class(tree1, tree2).compare(
            specific_files=['e'])
        self.assertEqual([], d.modified)
        # now give it an additional lookup:
        d = self.intertree_class(tree1, tree2).compare(
            specific_files=['e'], extra_trees=[tree3])
        self.assertEqual([], d.added)
        self.assertEqual(
            [('b/c', 'file', False, True)],
            [(c.path[1], c.kind[1], c.changed_content, c.meta_modified()) for c in d.modified])
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)

    def test_require_versioned(self):
        # this does not quite robustly test, as it is passing in missing paths
        # rather than present-but-not-versioned paths. At the moment there is
        # no mechanism for managing the test trees (which are readonly) to
        # get present-but-not-versioned files for trees that can do that.
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        self.assertRaises(errors.PathsNotVersionedError,
                          self.intertree_class(tree1, tree2).compare,
                          specific_files=['d'],
                          require_versioned=True)

    def test_default_ignores_unversioned_files(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree1/a', 'tree1/c',
                         'tree2/a', 'tree2/b', 'tree2/c'])
        tree1.add(['a', 'c'], ids=[b'a-id', b'c-id'])
        tree2.add(['a', 'c'], ids=[b'a-id', b'c-id'])

        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        d = self.intertree_class(tree1, tree2).compare()
        self.assertEqual([], d.added)
        self.assertEqual([('a', 'file', True, False),
                          ('c', 'file', True, False)],
                         [(c.path[1], c.kind[1], c.changed_content, c.meta_modified())
                          for c in d.modified])
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)
        self.assertEqual([], d.unversioned)

    def test_unversioned_paths_in_tree(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree2/file', 'tree2/dir/'])
        if supports_symlinks(self.test_dir):
            os.symlink('target', 'tree2/link')
            links_supported = True
        else:
            links_supported = False
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)
        d = self.intertree_class(tree1, tree2).compare(want_unversioned=True)
        self.assertEqual([], d.added)
        self.assertEqual([], d.modified)
        self.assertEqual([], d.removed)
        self.assertEqual([], d.renamed)
        self.assertEqual([], d.unchanged)
        expected_unversioned = [('dir', 'directory'),
                                ('file', 'file')]
        if links_supported:
            expected_unversioned.append(('link', 'symlink'))
        self.assertEqual(
            expected_unversioned,
            [(c.path[1], c.kind[1]) for c in d.unversioned])


class TestIterChanges(TestCaseWithTwoTrees):
    """Test the comparison iterator"""

    def _make_abc_tree(self, tree):
        """setup an abc content tree."""
        files = ['a', 'b/', 'b/c']
        self.build_tree(files, line_endings='binary',
                        transport=tree.controldir.root_transport)
        tree.set_root_id(b'root-id')
        tree.add(files, ids=[b'a-id', b'b-id', b'c-id'])

    def get_tree_no_parents_abc_content(self, tree, converter=None):
        """return a test tree with a, b/, b/c contents."""
        self._make_abc_tree(tree)
        return self._convert_tree(tree, converter)

    def get_tree_no_parents_abc_content_7(self, tree, converter=None):
        """return a test tree with a, b/, d/e contents.

        This variation adds a dir 'd' (b'd-id'), renames b to d/e.
        """
        self._make_abc_tree(tree)
        self.build_tree(['d/'], transport=tree.controldir.root_transport)
        tree.add(['d'], ids=[b'd-id'])
        tt = tree.transform()
        trans_id = tt.trans_id_tree_path('b')
        parent_trans_id = tt.trans_id_tree_path('d')
        tt.adjust_path('e', parent_trans_id, trans_id)
        tt.apply()
        return self._convert_tree(tree, converter)

    def assertEqualIterChanges(self, left_changes, right_changes):
        """Assert that left_changes == right_changes.

        :param left_changes: A list of the output from iter_changes.
        :param right_changes: A list of the output from iter_changes.
        """
        left_changes = self.sorted(left_changes)
        right_changes = self.sorted(right_changes)
        if left_changes == right_changes:
            return
        # setify to get item by item differences, but we can only do this
        # when all the ids are unique on both sides.
        left_dict = {item[0]: item for item in left_changes}
        right_dict = {item[0]: item for item in right_changes}
        if (len(left_dict) != len(left_changes)
                or len(right_dict) != len(right_changes)):
            # Can't do a direct comparison. We could do a sequence diff, but
            # for now just do a regular assertEqual for now.
            self.assertEqual(left_changes, right_changes)
        keys = set(left_dict).union(set(right_dict))
        different = []
        same = []
        for key in keys:
            left_item = left_dict.get(key)
            right_item = right_dict.get(key)
            if left_item == right_item:
                same.append(str(left_item))
            else:
                different.append(" {}\n {}".format(left_item, right_item))
        self.fail("iter_changes output different. Unchanged items:\n"
                  + "\n".join(same) + "\nChanged items:\n" + "\n".join(different))

    def do_iter_changes(self, tree1, tree2, **extra_args):
        """Helper to run iter_changes from tree1 to tree2.

        :param tree1, tree2:  The source and target trees. These will be locked
            automatically.
        :param **extra_args: Extra args to pass to iter_changes. This is not
            inspected by this test helper.
        """
        with tree1.lock_read(), tree2.lock_read():
            # sort order of output is not strictly defined
            return self.sorted(self.intertree_class(tree1, tree2)
                               .iter_changes(**extra_args))

    def check_has_changes(self, expected, tree1, tree2):
        # has_changes is defined for mutable trees only
        if not isinstance(tree2, mutabletree.MutableTree):
            if isinstance(tree1, mutabletree.MutableTree):
                # Let's switch the trees since has_changes() is commutative
                # (where we can apply it)
                tree2, tree1 = tree1, tree2
            else:
                # Neither tree can be used
                return
        with tree1.lock_read(), tree2.lock_read():
            return tree2.has_changes(tree1)

    def mutable_trees_to_locked_test_trees(self, tree1, tree2):
        """Convert the working trees into test trees.

        Read lock them, and add the unlock to the cleanup.
        """
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        tree1.lock_read()
        self.addCleanup(tree1.unlock)
        tree2.lock_read()
        self.addCleanup(tree2.unlock)
        return tree1, tree2

    def make_tree_with_special_names(self):
        """Create a tree with filenames chosen to exercise the walk order."""
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        paths = self._create_special_names(tree2, 'tree2')
        tree2.commit('initial', rev_id=b'rev-1')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        return (tree1, tree2, paths)

    def make_trees_with_special_names(self):
        """Both trees will use the special names.

        But the contents will differ for each file.
        """
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        paths = self._create_special_names(tree1, 'tree1')
        paths = self._create_special_names(tree2, 'tree2')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        return (tree1, tree2, paths)

    def _create_special_names(self, tree, base_path):
        """Create a tree with paths that expose differences in sort orders."""
        # Each directory will have a single file named 'f' inside
        dirs = ['a',
                'a-a',
                'a/a',
                'a/a-a',
                'a/a/a',
                'a/a/a-a',
                'a/a/a/a',
                'a/a/a/a-a',
                'a/a/a/a/a',
                ]
        with_slashes = []
        paths = []
        path_ids = []
        for d in dirs:
            with_slashes.append(base_path + '/' + d + '/')
            with_slashes.append(base_path + '/' + d + '/f')
            paths.append(d)
            paths.append(d + '/f')
            path_ids.append((d.replace('/', '_') + '-id').encode('ascii'))
            path_ids.append((d.replace('/', '_') + '_f-id').encode('ascii'))
        self.build_tree(with_slashes)
        tree.add(paths, ids=path_ids)
        return paths

    def test_compare_empty_trees(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_no_content(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        self.assertEqual([], self.do_iter_changes(tree1, tree2))
        self.check_has_changes(False, tree1, tree2)

    def added(self, tree, path):
        entry = self.get_path_entry(tree, path)
        return InventoryTreeChange(
            entry.file_id, (None, path), True, (False, True), (None, entry.parent_id),
            (None, entry.name), (None, entry.kind),
            (None, entry.executable))

    @staticmethod
    def get_path_entry(tree, path):
        iterator = tree.iter_entries_by_dir(specific_files=[path])
        try:
            return next(iterator)[1]
        except StopIteration:
            raise KeyError(path)

    def changed_content(self, tree, path):
        entry = self.get_path_entry(tree, path)
        return InventoryTreeChange(
            entry.file_id, (path, path), True, (True, True),
            (entry.parent_id, entry.parent_id),
            (entry.name, entry.name), (entry.kind, entry.kind),
            (entry.executable, entry.executable))

    def kind_changed(self, from_tree, to_tree, from_path, to_path):
        old_entry = self.get_path_entry(from_tree, from_path)
        new_entry = self.get_path_entry(to_tree, to_path)
        return InventoryTreeChange(
            new_entry.file_id, (from_path, to_path), True, (True, True),
            (old_entry.parent_id, new_entry.parent_id),
            (old_entry.name, new_entry.name),
            (old_entry.kind, new_entry.kind),
            (old_entry.executable, new_entry.executable))

    def missing(self, file_id, from_path, to_path, parent_id, kind):
        _, from_basename = os.path.split(from_path)
        _, to_basename = os.path.split(to_path)
        # missing files have both paths, but no kind.
        return InventoryTreeChange(
            file_id, (from_path, to_path), True, (True, True),
            (parent_id, parent_id),
            (from_basename, to_basename), (kind, None), (False, False))

    def deleted(self, tree, path):
        entry = self.get_path_entry(tree, path)
        return InventoryTreeChange(
            entry.file_id, (path, None), True, (True, False), (entry.parent_id, None),
            (entry.name, None), (entry.kind, None),
            (entry.executable, None))

    def renamed(self, from_tree, to_tree, from_path, to_path, content_changed):
        from_entry = self.get_path_entry(from_tree, from_path)
        to_entry = self.get_path_entry(to_tree, to_path)
        return InventoryTreeChange(
            to_entry.file_id, (from_path, to_path), content_changed, (True, True),
            (from_entry.parent_id, to_entry.parent_id),
            (from_entry.name, to_entry.name),
            (from_entry.kind, to_entry.kind),
            (from_entry.executable, to_entry.executable))

    def unchanged(self, tree, path):
        entry = self.get_path_entry(tree, path)
        parent = entry.parent_id
        name = entry.name
        kind = entry.kind
        executable = entry.executable
        return InventoryTreeChange(
            entry.file_id, (path, path), False, (True, True),
            (parent, parent), (name, name), (kind, kind),
            (executable, executable))

    def unversioned(self, tree, path):
        """Create an unversioned result."""
        _, basename = os.path.split(path)
        kind = tree._comparison_data(None, path)[0]
        return InventoryTreeChange(
            None, (None, path), True, (False, False), (None, None),
            (None, basename), (None, kind),
            (None, False))

    def sorted(self, changes):
        return sorted(changes, key=_change_key)

    def test_empty_to_abc_content(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        expected_results = self.sorted([
            self.added(tree2, ''),
            self.added(tree2, 'a'),
            self.added(tree2, 'b'),
            self.added(tree2, 'b/c'),
            self.deleted(tree1, '')])
        self.assertEqual(expected_results, self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_empty_specific_files(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.assertEqual([],
                         self.do_iter_changes(tree1, tree2, specific_files=[]))

    def test_no_specific_files(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        expected_results = self.sorted([
            self.added(tree2, ''),
            self.added(tree2, 'a'),
            self.added(tree2, 'b'),
            self.added(tree2, 'b/c'),
            self.deleted(tree1, '')])
        self.assertEqual(expected_results, self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_empty_to_abc_content_a_only(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.assertEqual(
            self.sorted([self.added(tree2, ''),
                         self.added(tree2, 'a'),
                         self.deleted(tree1, '')]),
            self.do_iter_changes(tree1, tree2, specific_files=['a']))

    def test_abc_content_to_empty_a_only(self):
        # For deletes we don't need to pickup parents.
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_no_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.assertEqual(
            [self.deleted(tree1, 'a')],
            self.do_iter_changes(tree1, tree2, specific_files=['a']))

    def test_abc_content_to_empty_b_only(self):
        # When b stops being a directory we have to pick up b/c as well.
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_no_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.assertEqual(
            [self.deleted(tree1, 'b'), self.deleted(tree1, 'b/c')],
            self.do_iter_changes(tree1, tree2, specific_files=['b']))

    def test_empty_to_abc_content_a_and_c_only(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_no_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        expected_result = self.sorted([
            self.added(tree2, ''),
            self.added(tree2, 'a'), self.added(tree2, 'b'),
            self.added(tree2, 'b/c'), self.deleted(tree1, '')])
        self.assertEqual(expected_result,
                         self.do_iter_changes(tree1, tree2, specific_files=['a', 'b/c']))

    def test_abc_content_to_empty(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_no_content(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        expected_results = self.sorted([
            self.added(tree2, ''),
            self.deleted(tree1, ''), self.deleted(tree1, 'a'),
            self.deleted(tree1, 'b'), self.deleted(tree1, 'b/c')])
        self.assertEqual(
            expected_results,
            self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_content_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_2(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        root_id = tree1.path2id('')
        self.assertEqual([(b'a-id', ('a', 'a'), True, (True, True),
                           (root_id, root_id), ('a', 'a'),
                           ('file', 'file'), (False, False), False)],
                         self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_meta_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_3(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        self.assertEqual([(b'c-id', ('b/c', 'b/c'), False, (True, True),
                           (b'b-id', b'b-id'), ('c', 'c'), ('file', 'file'),
                           (False, True), False)],
                         self.do_iter_changes(tree1, tree2))

    def test_empty_dir(self):
        """an empty dir should not cause glitches to surrounding files."""
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        # the pathname is chosen to fall between 'a' and 'b'.
        self.build_tree(['1/a-empty/', '2/a-empty/'])
        tree1.add(['a-empty'], ids=[b'a-empty'])
        tree2.add(['a-empty'], ids=[b'a-empty'])
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        expected = []
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))

    def test_file_rename(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_4(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        root_id = tree1.path2id('')
        self.assertEqual([(tree1.path2id('a'), ('a', 'd'), False, (True, True),
                           (root_id, root_id), ('a', 'd'), ('file', 'file'),
                           (False, False), False)],
                         self.do_iter_changes(tree1, tree2))

    def test_file_rename_and_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_5(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        root_id = tree1.path2id('')
        self.assertEqual([(b'a-id', ('a', 'd'), True, (True, True),
                           (root_id, root_id), ('a', 'd'), ('file', 'file'),
                           (False, False), False)],
                         self.do_iter_changes(tree1, tree2))

    def test_specific_content_modification_grabs_parents(self):
        # WHen the only direct change to a specified file is a content change,
        # and its in a reparented subtree, the parents are grabbed.
        tree1 = self.make_branch_and_tree('1')
        tree1.mkdir('changing', b'parent-id')
        tree1.mkdir('changing/unchanging', b'mid-id')
        tree1.add(['changing/unchanging/file'], ['file'], [b'file-id'])
        tree1.put_file_bytes_non_atomic(
            'changing/unchanging/file', b'a file')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree2.mkdir('changed', b'parent-id')
        tree2.mkdir('changed/unchanging', b'mid-id')
        tree2.add(['changed/unchanging/file'], ['file'], [b'file-id'])
        tree2.put_file_bytes_non_atomic(
            'changed/unchanging/file', b'changed content')
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        # parent-id has changed, as has file-id
        root_id = tree1.path2id('')
        self.assertEqualIterChanges(
            [self.renamed(tree1, tree2, 'changing', 'changed', False),
             self.renamed(tree1, tree2, 'changing/unchanging/file', 'changed/unchanging/file', True)],
            self.do_iter_changes(tree1, tree2,
                                 specific_files=['changed/unchanging/file']))

    def test_specific_content_modification_grabs_parents_root_changes(self):
        # WHen the only direct change to a specified file is a content change,
        # and its in a reparented subtree, the parents are grabbed, even if
        # that includes the root.
        tree1 = self.make_branch_and_tree('1')
        tree1.set_root_id(b'old')
        tree1.mkdir('changed', b'parent-id')
        tree1.mkdir('changed/unchanging', b'mid-id')
        tree1.add(['changed/unchanging/file'], ['file'], [b'file-id'])
        tree1.put_file_bytes_non_atomic('changed/unchanging/file', b'a file')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(b'new')
        tree2.mkdir('changed', b'parent-id')
        tree2.mkdir('changed/unchanging', b'mid-id')
        tree2.add(['changed/unchanging/file'], ['file'], [b'file-id'])
        tree2.put_file_bytes_non_atomic(
            'changed/unchanging/file', b'changed content')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # old is gone, new is added, parent-id has changed(reparented), as has
        # file-id(content)
        root_id = tree1.path2id('')
        self.assertEqualIterChanges(
            [self.renamed(tree1, tree2, 'changed', 'changed', False),
             self.added(tree2, ''),
             self.deleted(tree1, ''),
             self.renamed(tree1, tree2, 'changed/unchanging/file', 'changed/unchanging/file', True)],
            self.do_iter_changes(tree1, tree2,
                                 specific_files=['changed/unchanging/file']))

    def test_specific_with_rename_under_new_dir_reports_new_dir(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_7(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        # d(d-id) is new, e is b-id renamed.
        root_id = tree1.path2id('')
        self.assertEqualIterChanges(
            [self.renamed(tree1, tree2, 'b', 'd/e', False),
             self.added(tree2, 'd')],
            self.do_iter_changes(tree1, tree2, specific_files=['d/e']))

    def test_specific_with_rename_under_dir_under_new_dir_reports_new_dir(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_7(tree2)
        tree2.rename_one('a', 'd/e/a')
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        # d is new, d/e is b-id renamed, d/e/a is a-id renamed
        root_id = tree1.path2id('')
        self.assertEqualIterChanges(
            [self.renamed(tree1, tree2, 'b', 'd/e', False),
             self.added(tree2, 'd'),
             self.renamed(tree1, tree2, 'a', 'd/e/a', False)],
            self.do_iter_changes(tree1, tree2, specific_files=['d/e/a']))

    def test_specific_old_parent_same_path_new_parent(self):
        # when a parent is new at its path, if the path was used in the source
        # it must be emitted as a change.
        tree1 = self.make_branch_and_tree('1')
        tree1.add(['a'], ['file'], [b'a-id'])
        tree1.put_file_bytes_non_atomic('a', b'a file')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree2.mkdir('a', b'b-id')
        tree2.add(['a/c'], ['file'], [b'c-id'])
        tree2.put_file_bytes_non_atomic('a/c', b'another file')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # a-id is gone, b-id and c-id are added.
        self.assertEqualIterChanges(
            [self.deleted(tree1, 'a'),
             self.added(tree2, 'a'),
             self.added(tree2, 'a/c')],
            self.do_iter_changes(tree1, tree2, specific_files=['a/c']))

    def test_specific_old_parent_becomes_file(self):
        # When an old parent included because of a path conflict becomes a
        # non-directory, its children have to be all included in the delta.
        tree1 = self.make_branch_and_tree('1')
        tree1.mkdir('a', b'a-old-id')
        tree1.mkdir('a/reparented', b'reparented-id')
        tree1.mkdir('a/deleted', b'deleted-id')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree2.mkdir('a', b'a-new-id')
        tree2.mkdir('a/reparented', b'reparented-id')
        tree2.add(['b'], ['file'], [b'a-old-id'])
        tree2.put_file_bytes_non_atomic('b', b'')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # a-old-id is kind-changed, a-new-id is added, reparented-id is renamed,
        # deleted-id is gone
        self.assertEqualIterChanges(
            [self.kind_changed(tree1, tree2, 'a', 'b'),
             self.added(tree2, 'a'),
             self.renamed(tree1, tree2, 'a/reparented', 'a/reparented', False),
             self.deleted(tree1, 'a/deleted')],
            self.do_iter_changes(tree1, tree2,
                                 specific_files=['a/reparented']))

    def test_specific_old_parent_is_deleted(self):
        # When an old parent included because of a path conflict is removed,
        # its children have to be all included in the delta.
        tree1 = self.make_branch_and_tree('1')
        tree1.mkdir('a', b'a-old-id')
        tree1.mkdir('a/reparented', b'reparented-id')
        tree1.mkdir('a/deleted', b'deleted-id')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree2.mkdir('a', b'a-new-id')
        tree2.mkdir('a/reparented', b'reparented-id')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # a-old-id is gone, a-new-id is added, reparented-id is renamed,
        # deleted-id is gone
        self.assertEqualIterChanges(
            [self.deleted(tree1, 'a'),
             self.added(tree2, 'a'),
             self.renamed(tree1, tree2, 'a/reparented', 'a/reparented', False),
             self.deleted(tree1, 'a/deleted')],
            self.do_iter_changes(tree1, tree2,
                                 specific_files=['a/reparented']))

    def test_specific_old_parent_child_collides_with_unselected_new(self):
        # When the child of an old parent because of a path conflict becomes a
        # path conflict with some unselected item in the source, that item also
        # needs to be included (because otherwise the output of applying the
        # delta to the source would have two items at that path).
        tree1 = self.make_branch_and_tree('1')
        tree1.mkdir('a', b'a-old-id')
        tree1.mkdir('a/reparented', b'reparented-id')
        tree1.mkdir('collides', b'collides-id')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree2.mkdir('a', b'a-new-id')
        tree2.mkdir('a/selected', b'selected-id')
        tree2.mkdir('collides', b'reparented-id')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # a-old-id is one, a-new-id is added, reparented-id is renamed,
        # collides-id is gone, selected-id is new.
        self.assertEqualIterChanges(
            [self.deleted(tree1, 'a'),
             self.added(tree2, 'a'),
             self.renamed(tree1, tree2, 'a/reparented', 'collides', False),
             self.deleted(tree1, 'collides'),
             self.added(tree2, 'a/selected')],
            self.do_iter_changes(tree1, tree2,
                                 specific_files=['a/selected']))

    def test_specific_old_parent_child_dir_stops_being_dir(self):
        # When the child of an old parent also stops being a directory, its
        # children must also be included. This test checks that downward
        # recursion is done appropriately by starting at a child of the root of
        # a deleted subtree (a/reparented), and checking that a sibling
        # directory (a/deleted) has its children included in the delta.
        tree1 = self.make_branch_and_tree('1')
        tree1.mkdir('a', b'a-old-id')
        tree1.mkdir('a/reparented', b'reparented-id-1')
        tree1.mkdir('a/deleted', b'deleted-id-1')
        tree1.mkdir('a/deleted/reparented', b'reparented-id-2')
        tree1.mkdir('a/deleted/deleted', b'deleted-id-2')
        tree2 = self.make_to_branch_and_tree('2')
        tree2.set_root_id(tree1.path2id(''))
        tree2.mkdir('a', b'a-new-id')
        tree2.mkdir('a/reparented', b'reparented-id-1')
        tree2.mkdir('reparented', b'reparented-id-2')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # a-old-id is gone, a-new-id is added, reparented-id-1, -2 are renamed,
        # deleted-id-1 and -2 are gone.
        self.assertEqualIterChanges(
            [self.deleted(tree1, 'a'),
             self.added(tree2, 'a'),
             self.renamed(tree1, tree2, 'a/reparented', 'a/reparented', False),
             self.renamed(tree1, tree2, 'a/deleted/reparented', 'reparented', False),
             self.deleted(tree1, 'a/deleted'),
             self.deleted(tree1, 'a/deleted/deleted')],
            self.do_iter_changes(tree1, tree2,
                                 specific_files=['a/reparented']))

    def test_file_rename_and_meta_modification(self):
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_6(tree2)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        root_id = tree1.path2id('')
        self.assertEqual([(b'c-id', ('b/c', 'e'), False, (True, True),
                           (b'b-id', root_id), ('c', 'e'), ('file', 'file'),
                           (False, True), False)],
                         self.do_iter_changes(tree1, tree2))

    def test_file_becomes_unversionable_bug_438569(self):
        # This isn't strictly a intertree problem, but its the intertree code
        # path that triggers all stat cache updates on both xml and dirstate
        # trees.
        # In bug 438569, a file becoming a fifo causes an assert. Fifo's are
        # not versionable or diffable. For now, we simply stop cold when they
        # are detected (because we don't know how far through the code the
        # assumption 'fifo's do not exist' goes). In future we could report
        # the kind change and have commit refuse to go futher, or something
        # similar. One particular reason for choosing this approach is that
        # there is no minikind for 'fifo' in dirstate today, so we can't
        # actually update records that way.
        # To add confusion, the totally generic code path works - but it
        # doesn't update persistent metadata. So this test permits InterTrees
        # to either work, or fail with BadFileKindError.
        self.requireFeature(features.OsFifoFeature)
        tree1 = self.make_branch_and_tree('1')
        self.build_tree(['1/a'])
        tree1.set_root_id(b'root-id')
        tree1.add(['a'], ids=[b'a-id'])
        tree2 = self.make_branch_and_tree('2')
        os.mkfifo('2/a')
        tree2.add(['a'], ['file'], [b'a-id'])
        try:
            tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        except (KeyError,):
            raise tests.TestNotApplicable(
                "Cannot represent a FIFO in this case %s" % self.id())
        try:
            self.do_iter_changes(tree1, tree2)
        except errors.BadFileKindError:
            pass

    def test_missing_in_target(self):
        """Test with the target files versioned but absent from disk."""
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content(tree2)
        os.unlink('2/a')
        shutil.rmtree('2/b')
        # TODO ? have a symlink here?
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        self.not_applicable_if_missing_in('a', tree2)
        self.not_applicable_if_missing_in('b', tree2)
        expected = self.sorted([
            self.missing(b'a-id', 'a', 'a', b'root-id', 'file'),
            self.missing(b'b-id', 'b', 'b', b'root-id', 'directory'),
            self.missing(b'c-id', 'b/c', 'b/c', b'b-id', 'file'),
            ])
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))

    def test_missing_and_renamed(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree1/file'])
        tree1.add(['file'], ids=[b'file-id'])
        self.build_tree(['tree2/directory/'])
        tree2.add(['directory'], ids=[b'file-id'])
        os.rmdir('tree2/directory')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_missing_in('directory', tree2)

        root_id = tree1.path2id('')
        expected = self.sorted([
            self.missing(b'file-id', 'file', 'directory', root_id, 'file'),
            ])
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))

    def test_only_in_source_and_missing(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree1/file'])
        tree1.add(['file'], ids=[b'file-id'])
        os.unlink('tree1/file')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_missing_in('file', tree1)
        root_id = tree1.path2id('')
        expected = [
            InventoryTreeChange(
                b'file-id', ('file', None), False, (True, False),
                (root_id, None), ('file', None), (None, None), (False, None))]
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))

    def test_only_in_target_and_missing(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree2/file'])
        tree2.add(['file'], ids=[b'file-id'])
        os.unlink('tree2/file')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_missing_in('file', tree2)
        root_id = tree1.path2id('')
        expected = [
            InventoryTreeChange(
                b'file-id', (None, 'file'), False, (False, True),
                (None, root_id), (None, 'file'), (None, None), (None, False))]
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))

    def test_only_in_target_missing_subtree_specific_bug_367632(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree2/a-dir/', 'tree2/a-dir/a-file'])
        tree2.add(['a-dir', 'a-dir/a-file'], ids=[b'dir-id', b'file-id'])
        os.unlink('tree2/a-dir/a-file')
        os.rmdir('tree2/a-dir')
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_missing_in('a-dir', tree2)
        root_id = tree1.path2id('')
        expected = [
            InventoryTreeChange(
                b'dir-id', (None, 'a-dir'), False, (False, True),
                (None, root_id), (None, 'a-dir'), (None, None), (None, False)),
            InventoryTreeChange(
                b'file-id', (None, 'a-dir/a-file'), False, (False, True),
                (None, b'dir-id'), (None, 'a-file'), (None, None), (None, False))
            ]
        # bug 367632 showed that specifying the root broke some code paths,
        # so we check this contract with and without it.
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2, specific_files=['']))

    def test_unchanged_with_renames_and_modifications(self):
        """want_unchanged should generate a list of unchanged entries."""
        tree1 = self.make_branch_and_tree('1')
        tree2 = self.make_to_branch_and_tree('2')
        tree1 = self.get_tree_no_parents_abc_content(tree1)
        tree2 = self.get_tree_no_parents_abc_content_5(tree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.assertEqual(sorted([self.unchanged(tree1, ''),
                                 self.unchanged(tree1, 'b'),
                                 InventoryTreeChange(
                                     b'a-id', ('a', 'd'), True, (True, True),
                                     (b'root-id', b'root-id'), ('a', 'd'),
                                     ('file', 'file'),
                                     (False, False)), self.unchanged(tree1, 'b/c')]),
                         self.do_iter_changes(tree1, tree2, include_unchanged=True))

    def test_compare_subtrees(self):
        tree1 = self.make_branch_and_tree('1')
        if not tree1.supports_tree_reference():
            return
        tree1.set_root_id(b'root-id')
        subtree1 = self.make_branch_and_tree('1/sub')
        subtree1.set_root_id(b'subtree-id')
        tree1.add_reference(subtree1)

        tree2 = self.make_to_branch_and_tree('2')
        if not tree2.supports_tree_reference():
            return
        tree2.set_root_id(b'root-id')
        subtree2 = self.make_to_branch_and_tree('2/sub')
        subtree2.set_root_id(b'subtree-id')
        tree2.add_reference(subtree2)
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        self.assertEqual([], list(tree2.iter_changes(tree1)))
        subtree1.commit('commit', rev_id=b'commit-a')
        self.assertThat(
            tree2.iter_changes(tree1, include_unchanged=True),
            MatchesTreeChanges(
                tree1, tree2, [
                    TreeChange(
                        ('', ''),
                        False,
                        (True, True),
                        ('', ''),
                        ('directory', 'directory'),
                        (False, False)),
                    TreeChange(
                        ('sub', 'sub',),
                        False,
                        (True, True),
                        ('sub', 'sub'),
                        ('tree-reference', 'tree-reference'),
                        (False, False))],
                    ))

    def test_disk_in_subtrees_skipped(self):
        """subtrees are considered not-in-the-current-tree.

        This test tests the trivial case, where the basis has no paths in the
        current trees subtree.
        """
        tree1 = self.make_branch_and_tree('1')
        tree1.set_root_id(b'root-id')
        tree2 = self.make_to_branch_and_tree('2')
        if not tree2.supports_tree_reference():
            return
        tree2.set_root_id(b'root-id')
        subtree2 = self.make_to_branch_and_tree('2/sub')
        subtree2.set_root_id(b'subtree-id')
        tree2.add_reference(subtree2)
        self.build_tree(['2/sub/file'])
        subtree2.add(['file'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # this should filter correctly from above
        self.assertEqual([self.added(tree2, 'sub')],
                         self.do_iter_changes(tree1, tree2, want_unversioned=True))
        # and when the path is named
        self.assertEqual([self.added(tree2, 'sub')],
                         self.do_iter_changes(tree1, tree2, specific_files=['sub'],
                                              want_unversioned=True))

    def test_default_ignores_unversioned_files(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree1/a', 'tree1/c',
                         'tree2/a', 'tree2/b', 'tree2/c'])
        tree1.add(['a', 'c'], ids=[b'a-id', b'c-id'])
        tree2.add(['a', 'c'], ids=[b'a-id', b'c-id'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        # We should ignore the fact that 'b' exists in tree-2
        # because the want_unversioned parameter was not given.
        expected = self.sorted([
            self.changed_content(tree2, 'a'),
            self.changed_content(tree2, 'c'),
            ])
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_unversioned_paths_in_tree(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree2/file', 'tree2/dir/'])
        if supports_symlinks(self.test_dir):
            os.symlink('target', 'tree2/link')
            links_supported = True
        else:
            links_supported = False
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)
        expected = [
            self.unversioned(tree2, 'file'),
            self.unversioned(tree2, 'dir'),
            ]
        if links_supported:
            expected.append(self.unversioned(tree2, 'link'))
        expected = self.sorted(expected)
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2,
                                                        want_unversioned=True))

    def test_unversioned_paths_in_tree_specific_files(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        self.build_tree(['tree2/file', 'tree2/dir/'])
        if supports_symlinks(self.test_dir):
            os.symlink('target', 'tree2/link')
            links_supported = True
        else:
            links_supported = False
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)
        expected = [
            self.unversioned(tree2, 'file'),
            self.unversioned(tree2, 'dir'),
            ]
        specific_files = ['file', 'dir']
        if links_supported:
            expected.append(self.unversioned(tree2, 'link'))
            specific_files.append('link')
        expected = self.sorted(expected)
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2,
                                                        specific_files=specific_files, require_versioned=False,
                                                        want_unversioned=True))

    def test_unversioned_paths_in_target_matching_source_old_names(self):
        # its likely that naive implementations of unversioned file support
        # will fail if the path was versioned, but is not any more,
        # due to a rename, not due to unversioning it.
        # That is, if the old tree has a versioned file 'foo', and
        # the new tree has the same file but versioned as 'bar', and also
        # has an unknown file 'foo', we should get back output for
        # both foo and bar.
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree2/file', 'tree2/dir/',
                         'tree1/file', 'tree2/movedfile',
                         'tree1/dir/', 'tree2/moveddir/'])
        if supports_symlinks(self.test_dir):
            os.symlink('target', 'tree1/link')
            os.symlink('target', 'tree2/link')
            os.symlink('target', 'tree2/movedlink')
            links_supported = True
        else:
            links_supported = False
        tree1.add(['file', 'dir'], ids=[b'file-id', b'dir-id'])
        tree2.add(['movedfile', 'moveddir'], ids=[b'file-id', b'dir-id'])
        if links_supported:
            tree1.add(['link'], ids=[b'link-id'])
            tree2.add(['movedlink'], ids=[b'link-id'])
        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)
        root_id = tree1.path2id('')
        expected = [
            self.renamed(tree1, tree2, 'dir', 'moveddir', False),
            self.renamed(tree1, tree2, 'file', 'movedfile', True),
            self.unversioned(tree2, 'file'),
            self.unversioned(tree2, 'dir'),
            ]
        specific_files = ['file', 'dir']
        if links_supported:
            expected.append(self.renamed(tree1, tree2, 'link', 'movedlink', False))
            expected.append(self.unversioned(tree2, 'link'))
            specific_files.append('link')
        expected = self.sorted(expected)
        # run once with, and once without specific files, to catch
        # potentially different code paths.
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2,
                                                        require_versioned=False,
                                                        want_unversioned=True))
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2,
                                                        specific_files=specific_files, require_versioned=False,
                                                        want_unversioned=True))

    def test_similar_filenames(self):
        """Test when we have a few files with similar names."""
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))

        # The trees are actually identical, but they happen to contain
        # similarly named files.
        self.build_tree(['tree1/a/',
                         'tree1/a/b/',
                         'tree1/a/b/c/',
                         'tree1/a/b/c/d/',
                         'tree1/a-c/',
                         'tree1/a-c/e/',
                         'tree2/a/',
                         'tree2/a/b/',
                         'tree2/a/b/c/',
                         'tree2/a/b/c/d/',
                         'tree2/a-c/',
                         'tree2/a-c/e/',
                         ])
        tree1.add(['a', 'a/b', 'a/b/c', 'a/b/c/d', 'a-c', 'a-c/e'],
                  ids=[b'a-id', b'b-id', b'c-id', b'd-id', b'a-c-id', b'e-id'])
        tree2.add(['a', 'a/b', 'a/b/c', 'a/b/c/d', 'a-c', 'a-c/e'],
                  ids=[b'a-id', b'b-id', b'c-id', b'd-id', b'a-c-id', b'e-id'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)

        self.assertEqual([], self.do_iter_changes(tree1, tree2,
                                                  want_unversioned=True))
        expected = self.sorted([
            self.unchanged(tree2, ''),
            self.unchanged(tree2, 'a'),
            self.unchanged(tree2, 'a/b'),
            self.unchanged(tree2, 'a/b/c'),
            self.unchanged(tree2, 'a/b/c/d'),
            self.unchanged(tree2, 'a-c'),
            self.unchanged(tree2, 'a-c/e'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              want_unversioned=True,
                                              include_unchanged=True))

    def test_unversioned_subtree_only_emits_root(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree2/dir/', 'tree2/dir/file'])
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)
        expected = [
            self.unversioned(tree2, 'dir'),
            ]
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2,
                                                        want_unversioned=True))

    def make_trees_with_symlinks(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree1/fromfile', 'tree1/fromdir/'])
        self.build_tree(['tree2/tofile', 'tree2/todir/', 'tree2/unknown'])
        os.symlink('original', 'tree1/changed')
        os.symlink('original', 'tree1/removed')
        os.symlink('original', 'tree1/tofile')
        os.symlink('original', 'tree1/todir')
        # we make the unchanged link point at unknown to catch incorrect
        # symlink-following code in the specified_files test.
        os.symlink('unknown', 'tree1/unchanged')
        os.symlink('new', 'tree2/added')
        os.symlink('new', 'tree2/changed')
        os.symlink('new', 'tree2/fromfile')
        os.symlink('new', 'tree2/fromdir')
        os.symlink('unknown', 'tree2/unchanged')
        from_paths_and_ids = [
            'fromdir',
            'fromfile',
            'changed',
            'removed',
            'todir',
            'tofile',
            'unchanged',
            ]
        to_paths_and_ids = [
            'added',
            'fromdir',
            'fromfile',
            'changed',
            'todir',
            'tofile',
            'unchanged',
            ]
        tree1.add(from_paths_and_ids, ids=[p.encode('utf-8')
                                       for p in from_paths_and_ids])
        tree2.add(to_paths_and_ids, ids=[p.encode('utf-8')
                                     for p in to_paths_and_ids])
        return self.mutable_trees_to_locked_test_trees(tree1, tree2)

    def test_versioned_symlinks(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree1, tree2 = self.make_trees_with_symlinks()
        self.not_applicable_if_cannot_represent_unversioned(tree2)
        root_id = tree1.path2id('')
        expected = [
            self.unchanged(tree1, ''),
            self.added(tree2, 'added'),
            self.changed_content(tree2, 'changed'),
            self.kind_changed(tree1, tree2, 'fromdir', 'fromdir'),
            self.kind_changed(tree1, tree2, 'fromfile', 'fromfile'),
            self.deleted(tree1, 'removed'),
            self.unchanged(tree2, 'unchanged'),
            self.unversioned(tree2, 'unknown'),
            self.kind_changed(tree1, tree2, 'todir', 'todir'),
            self.kind_changed(tree1, tree2, 'tofile', 'tofile'),
            ]
        expected = self.sorted(expected)
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2, include_unchanged=True,
                                              want_unversioned=True))
        self.check_has_changes(True, tree1, tree2)

    def test_versioned_symlinks_specific_files(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree1, tree2 = self.make_trees_with_symlinks()
        root_id = tree1.path2id('')
        expected = [
            self.added(tree2, 'added'),
            self.changed_content(tree2, 'changed'),
            self.kind_changed(tree1, tree2, 'fromdir', 'fromdir'),
            self.kind_changed(tree1, tree2, 'fromfile', 'fromfile'),
            self.deleted(tree1, 'removed'),
            self.kind_changed(tree1, tree2, 'todir', 'todir'),
            self.kind_changed(tree1, tree2, 'tofile', 'tofile'),
            ]
        expected = self.sorted(expected)
        # we should get back just the changed links. We pass in 'unchanged' to
        # make sure that it is correctly not returned - and neither is the
        # unknown path 'unknown' which it points at.
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2,
                                                        specific_files=['added', 'changed', 'fromdir', 'fromfile',
                                                                        'removed', 'unchanged', 'todir', 'tofile']))
        self.check_has_changes(True, tree1, tree2)

    def test_tree_with_special_names(self):
        tree1, tree2, paths = self.make_tree_with_special_names()
        expected = self.sorted(self.added(tree2, p) for p in paths)
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_trees_with_special_names(self):
        tree1, tree2, paths = self.make_trees_with_special_names()
        expected = self.sorted(self.changed_content(tree2, p) for p in paths
                               if p.endswith('/f'))
        self.assertEqual(expected, self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_trees_with_deleted_dir(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        self.build_tree(['tree1/a', 'tree1/b/', 'tree1/b/c',
                         'tree1/b/d/', 'tree1/b/d/e', 'tree1/f/', 'tree1/f/g',
                         'tree2/a', 'tree2/f/', 'tree2/f/g'])
        tree1.add(['a', 'b', 'b/c', 'b/d/', 'b/d/e', 'f', 'f/g'],
                  ids=[b'a-id', b'b-id', b'c-id', b'd-id', b'e-id', b'f-id', b'g-id'])
        tree2.add(['a', 'f', 'f/g'], ids=[b'a-id', b'f-id', b'g-id'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        # We should notice that 'b' and all its children are deleted
        expected = [
            self.changed_content(tree2, 'a'),
            self.changed_content(tree2, 'f/g'),
            self.deleted(tree1, 'b'),
            self.deleted(tree1, 'b/c'),
            self.deleted(tree1, 'b/d'),
            self.deleted(tree1, 'b/d/e'),
            ]
        self.assertEqualIterChanges(expected,
                                    self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_added_unicode(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # u'\u03b1' == GREEK SMALL LETTER ALPHA
        # u'\u03c9' == GREEK SMALL LETTER OMEGA
        a_id = '\u03b1-id'.encode()
        added_id = '\u03c9_added_id'.encode()
        added_path = '\u03b1/\u03c9-added'
        try:
            self.build_tree(['tree1/\u03b1/',
                             'tree2/\u03b1/',
                             'tree2/\u03b1/\u03c9-added',
                             ])
        except UnicodeError:
            raise tests.TestSkipped("Could not create Unicode files.")
        tree1.add(['\u03b1'], ids=[a_id])
        tree2.add(['\u03b1', '\u03b1/\u03c9-added'], ids=[a_id, added_id])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        self.assertEqual([self.added(tree2, added_path)],
                         self.do_iter_changes(tree1, tree2))
        self.assertEqual([self.added(tree2, added_path)],
                         self.do_iter_changes(tree1, tree2,
                                              specific_files=['\u03b1']))
        self.check_has_changes(True, tree1, tree2)

    def test_deleted_unicode(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # u'\u03b1' == GREEK SMALL LETTER ALPHA
        # u'\u03c9' == GREEK SMALL LETTER OMEGA
        a_id = '\u03b1-id'.encode()
        deleted_id = '\u03c9_deleted_id'.encode()
        deleted_path = '\u03b1/\u03c9-deleted'
        try:
            self.build_tree(['tree1/\u03b1/',
                             'tree1/\u03b1/\u03c9-deleted',
                             'tree2/\u03b1/',
                             ])
        except UnicodeError:
            raise tests.TestSkipped("Could not create Unicode files.")
        tree1.add(['\u03b1', '\u03b1/\u03c9-deleted'], ids=[a_id, deleted_id])
        tree2.add(['\u03b1'], ids=[a_id])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        self.assertEqual([self.deleted(tree1, deleted_path)],
                         self.do_iter_changes(tree1, tree2))
        self.assertEqual([self.deleted(tree1, deleted_path)],
                         self.do_iter_changes(tree1, tree2,
                                              specific_files=['\u03b1']))
        self.check_has_changes(True, tree1, tree2)

    def test_modified_unicode(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # u'\u03b1' == GREEK SMALL LETTER ALPHA
        # u'\u03c9' == GREEK SMALL LETTER OMEGA
        a_id = '\u03b1-id'.encode()
        mod_id = '\u03c9_mod_id'.encode()
        mod_path = '\u03b1/\u03c9-modified'
        try:
            self.build_tree(['tree1/\u03b1/',
                             'tree1/' + mod_path,
                             'tree2/\u03b1/',
                             'tree2/' + mod_path,
                             ])
        except UnicodeError:
            raise tests.TestSkipped("Could not create Unicode files.")
        tree1.add(['\u03b1', mod_path], ids=[a_id, mod_id])
        tree2.add(['\u03b1', mod_path], ids=[a_id, mod_id])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        self.assertEqual([self.changed_content(tree1, mod_path)],
                         self.do_iter_changes(tree1, tree2))
        self.assertEqual([self.changed_content(tree1, mod_path)],
                         self.do_iter_changes(tree1, tree2,
                                              specific_files=['\u03b1']))
        self.check_has_changes(True, tree1, tree2)

    def test_renamed_unicode(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # u'\u03b1' == GREEK SMALL LETTER ALPHA
        # u'\u03c9' == GREEK SMALL LETTER OMEGA
        a_id = '\u03b1-id'.encode()
        rename_id = '\u03c9_rename_id'.encode()
        try:
            self.build_tree(['tree1/\u03b1/',
                             'tree2/\u03b1/',
                             ])
        except UnicodeError:
            raise tests.TestSkipped("Could not create Unicode files.")
        self.build_tree_contents([('tree1/\u03c9-source', b'contents\n'),
                                  ('tree2/\u03b1/\u03c9-target', b'contents\n'),
                                  ])
        tree1.add(['\u03b1', '\u03c9-source'], ids=[a_id, rename_id])
        tree2.add(['\u03b1', '\u03b1/\u03c9-target'], ids=[a_id, rename_id])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        self.assertEqual([self.renamed(tree1, tree2, '\u03c9-source', '\u03b1/\u03c9-target', False)],
                         self.do_iter_changes(tree1, tree2))
        self.assertEqualIterChanges(
            [self.renamed(tree1, tree2, '\u03c9-source', '\u03b1/\u03c9-target', False)],
            self.do_iter_changes(tree1, tree2, specific_files=['\u03b1']))
        self.check_has_changes(True, tree1, tree2)

    def test_unchanged_unicode(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        # u'\u03b1' == GREEK SMALL LETTER ALPHA
        # u'\u03c9' == GREEK SMALL LETTER OMEGA
        a_id = '\u03b1-id'.encode()
        subfile_id = '\u03c9-subfile-id'.encode()
        rootfile_id = '\u03c9-root-id'.encode()
        try:
            self.build_tree(['tree1/\u03b1/',
                             'tree2/\u03b1/',
                             ])
        except UnicodeError:
            raise tests.TestSkipped("Could not create Unicode files.")
        self.build_tree_contents([
            ('tree1/\u03b1/\u03c9-subfile', b'sub contents\n'),
            ('tree2/\u03b1/\u03c9-subfile', b'sub contents\n'),
            ('tree1/\u03c9-rootfile', b'root contents\n'),
            ('tree2/\u03c9-rootfile', b'root contents\n'),
            ])
        tree1.add(['\u03b1', '\u03b1/\u03c9-subfile', '\u03c9-rootfile'],
                  ids=[a_id, subfile_id, rootfile_id])
        tree2.add(['\u03b1', '\u03b1/\u03c9-subfile', '\u03c9-rootfile'],
                  ids=[a_id, subfile_id, rootfile_id])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        expected = self.sorted([
            self.unchanged(tree1, ''),
            self.unchanged(tree1, '\u03b1'),
            self.unchanged(tree1, '\u03b1/\u03c9-subfile'),
            self.unchanged(tree1, '\u03c9-rootfile'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              include_unchanged=True))

        # We should also be able to select just a subset
        expected = self.sorted([
            self.unchanged(tree1, '\u03b1'),
            self.unchanged(tree1, '\u03b1/\u03c9-subfile'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2, specific_files=['\u03b1'],
                                              include_unchanged=True))

    def test_unknown_unicode(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        tree2.set_root_id(tree1.path2id(''))
        # u'\u03b1' == GREEK SMALL LETTER ALPHA
        # u'\u03c9' == GREEK SMALL LETTER OMEGA
        a_id = '\u03b1-id'.encode()
        try:
            self.build_tree(['tree1/\u03b1/',
                             'tree2/\u03b1/',
                             'tree2/\u03b1/unknown_dir/',
                             'tree2/\u03b1/unknown_file',
                             'tree2/\u03b1/unknown_dir/file',
                             'tree2/\u03c9-unknown_root_file',
                             ])
        except UnicodeError:
            raise tests.TestSkipped("Could not create Unicode files.")
        tree1.add(['\u03b1'], ids=[a_id])
        tree2.add(['\u03b1'], ids=[a_id])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)

        expected = self.sorted([
            self.unversioned(tree2, '\u03b1/unknown_dir'),
            self.unversioned(tree2, '\u03b1/unknown_file'),
            self.unversioned(tree2, '\u03c9-unknown_root_file'),
            # a/unknown_dir/file should not be included because we should not
            # recurse into unknown_dir
            # self.unversioned(tree2, 'a/unknown_dir/file'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              require_versioned=False,
                                              want_unversioned=True))
        self.assertEqual([],  # Without want_unversioned we should get nothing
                         self.do_iter_changes(tree1, tree2))
        self.check_has_changes(False, tree1, tree2)

        # We should also be able to select just a subset
        expected = self.sorted([
            self.unversioned(tree2, '\u03b1/unknown_dir'),
            self.unversioned(tree2, '\u03b1/unknown_file'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              specific_files=['\u03b1'],
                                              require_versioned=False,
                                              want_unversioned=True))
        self.assertEqual([],  # Without want_unversioned we should get nothing
                         self.do_iter_changes(tree1, tree2,
                                              specific_files=['\u03b1']))

    def test_unknown_empty_dir(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # Start with 2 identical trees
        self.build_tree(['tree1/a/', 'tree1/b/',
                         'tree2/a/', 'tree2/b/'])
        self.build_tree_contents([('tree1/b/file', b'contents\n'),
                                  ('tree2/b/file', b'contents\n')])
        tree1.add(['a', 'b', 'b/file'], ids=[b'a-id', b'b-id', b'b-file-id'])
        tree2.add(['a', 'b', 'b/file'], ids=[b'a-id', b'b-id', b'b-file-id'])

        # Now create some unknowns in tree2
        # We should find both a/file and a/dir as unknown, but we shouldn't
        # recurse into a/dir to find that a/dir/subfile is also unknown.
        self.build_tree(
            ['tree2/a/file', 'tree2/a/dir/', 'tree2/a/dir/subfile'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)

        if tree2.has_versioned_directories():
            expected = self.sorted([
                self.unversioned(tree2, 'a/file'),
                self.unversioned(tree2, 'a/dir'),
                ])
            self.assertEqual(expected,
                             self.do_iter_changes(tree1, tree2,
                                                  require_versioned=False,
                                                  want_unversioned=True))
        else:
            expected = self.sorted([
                self.unversioned(tree2, 'a/file'),
                self.unversioned(tree2, 'a/dir/subfile'),
                ])
            self.assertEqual(expected,
                             self.do_iter_changes(tree1, tree2,
                                                  require_versioned=False,
                                                  want_unversioned=True))

    def test_rename_over_deleted(self):
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # The final changes should be:
        #   touch a b c d
        #   add a b c d
        #   commit
        #   rm a d
        #   mv b a
        #   mv c d
        self.build_tree_contents([
            ('tree1/a', b'a contents\n'),
            ('tree1/b', b'b contents\n'),
            ('tree1/c', b'c contents\n'),
            ('tree1/d', b'd contents\n'),
            ('tree2/a', b'b contents\n'),
            ('tree2/d', b'c contents\n'),
            ])
        tree1.add(['a', 'b', 'c', 'd'], ids=[b'a-id', b'b-id', b'c-id', b'd-id'])
        tree2.add(['a', 'd'], ids=[b'b-id', b'c-id'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        expected = self.sorted([
            self.deleted(tree1, 'a'),
            self.deleted(tree1, 'd'),
            self.renamed(tree1, tree2, 'b', 'a', False),
            self.renamed(tree1, tree2, 'c', 'd', False),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2))
        self.check_has_changes(True, tree1, tree2)

    def test_deleted_and_unknown(self):
        """Test a file marked removed, but still present on disk."""
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # The final changes should be:
        # bzr add a b c
        # bzr rm --keep b
        self.build_tree_contents([
            ('tree1/a', b'a contents\n'),
            ('tree1/b', b'b contents\n'),
            ('tree1/c', b'c contents\n'),
            ('tree2/a', b'a contents\n'),
            ('tree2/b', b'b contents\n'),
            ('tree2/c', b'c contents\n'),
            ])
        tree1.add(['a', 'b', 'c'], ids=[b'a-id', b'b-id', b'c-id'])
        tree2.add(['a', 'c'], ids=[b'a-id', b'c-id'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_cannot_represent_unversioned(tree2)

        expected = self.sorted([
            self.deleted(tree1, 'b'),
            self.unversioned(tree2, 'b'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              want_unversioned=True))
        expected = self.sorted([
            self.deleted(tree1, 'b'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              want_unversioned=False))

    def test_renamed_and_added(self):
        """Test when we have renamed a file, and put another in its place."""
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # The final changes are:
        # bzr add b c
        # bzr mv b a
        # bzr mv c d
        # bzr add b c

        self.build_tree_contents([
            ('tree1/b', b'b contents\n'),
            ('tree1/c', b'c contents\n'),
            ('tree2/a', b'b contents\n'),
            ('tree2/b', b'new b contents\n'),
            ('tree2/c', b'new c contents\n'),
            ('tree2/d', b'c contents\n'),
            ])
        tree1.add(['b', 'c'], ids=[b'b1-id', b'c1-id'])
        tree2.add(['a', 'b', 'c', 'd'], ids=[
                  b'b1-id', b'b2-id', b'c2-id', b'c1-id'])

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)

        expected = self.sorted([
            self.renamed(tree1, tree2, 'b', 'a', False),
            self.renamed(tree1, tree2, 'c', 'd', False),
            self.added(tree2, 'b'),
            self.added(tree2, 'c'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              want_unversioned=True))

    def test_renamed_and_unknown(self):
        """A file was moved on the filesystem, but not in bzr."""
        tree1 = self.make_branch_and_tree('tree1')
        tree2 = self.make_to_branch_and_tree('tree2')
        root_id = tree1.path2id('')
        tree2.set_root_id(root_id)

        # The final changes are:
        # bzr add a b
        # mv a a2

        self.build_tree_contents([
            ('tree1/a', b'a contents\n'),
            ('tree1/b', b'b contents\n'),
            ('tree2/a', b'a contents\n'),
            ('tree2/b', b'b contents\n'),
            ])
        tree1.add(['a', 'b'], ids=[b'a-id', b'b-id'])
        tree2.add(['a', 'b'], ids=[b'a-id', b'b-id'])
        os.rename('tree2/a', 'tree2/a2')

        tree1, tree2 = self.mutable_trees_to_locked_test_trees(tree1, tree2)
        self.not_applicable_if_missing_in('a', tree2)

        expected = self.sorted([
            self.missing(b'a-id', 'a', 'a', tree2.path2id(''), 'file'),
            self.unversioned(tree2, 'a2'),
            ])
        self.assertEqual(expected,
                         self.do_iter_changes(tree1, tree2,
                                              want_unversioned=True))



# Copyright (C) 2008-2018 Jelmer Vernooij <jelmer@jelmer.uk>
# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""An adapter between a Git Repository and a Breezy one."""

from io import BytesIO

from dulwich.errors import NotCommitError
from dulwich.object_store import peel_sha, tree_lookup_path
from dulwich.objects import ZERO_SHA, Commit

from .. import check, errors
from .. import graph as _mod_graph
from .. import lock, repository
from .. import revision as _mod_revision
from .. import trace, transactions, ui
from ..decorators import only_raises
from ..foreign import ForeignRepository
from .filegraph import GitFileLastChangeScanner, GitFileParentProvider
from .mapping import (default_mapping, encode_git_path, foreign_vcs_git,
                      mapping_registry)
from .tree import GitRevisionTree


class GitCheck(check.Check):

    def __init__(self, repository, check_repo=True):
        self.repository = repository
        self.check_repo = check_repo
        self.checked_rev_cnt = 0
        self.object_count = None
        self.problems = []

    def check(self, callback_refs=None, check_repo=True):
        if callback_refs is None:
            callback_refs = {}
        with self.repository.lock_read(), \
                ui.ui_factory.nested_progress_bar() as self.progress:
            shas = set(self.repository._git.object_store)
            self.object_count = len(shas)
            # TODO(jelmer): Check more things
            for i, sha in enumerate(shas):
                self.progress.update('checking objects', i, self.object_count)
                o = self.repository._git.object_store[sha]
                try:
                    o.check()
                except Exception as e:
                    self.problems.append((sha, e))

    def _report_repo_results(self, verbose):
        trace.note('checked repository {} format {}'.format(
            self.repository.user_url,
            self.repository._format))
        trace.note('%6d objects', self.object_count)
        for sha, problem in self.problems:
            trace.note('%s: %s', sha, problem)

    def report_results(self, verbose):
        if self.check_repo:
            self._report_repo_results(verbose)


for optimiser in ['InterRemoteGitNonGitRepository',
                  'InterLocalGitNonGitRepository',
                  'InterLocalGitLocalGitRepository',
                  'InterLocalGitRemoteGitRepository',
                  'InterRemoteGitLocalGitRepository',
                  'InterToLocalGitRepository',
                  'InterToRemoteGitRepository',
                  ]:
    repository.InterRepository.register_lazy_optimiser(
        'breezy.git.interrepo', optimiser)


class GitRepository(ForeignRepository):
    """An adapter to git repositories for bzr."""

    vcs = foreign_vcs_git
    chk_bytes = None

    def __init__(self, gitdir):
        self._transport = gitdir.root_transport
        super().__init__(GitRepositoryFormat(),
                                            gitdir, control_files=None)
        self.base = gitdir.root_transport.base
        self._lock_mode = None
        self._lock_count = 0

    def add_fallback_repository(self, basis_url):
        raise errors.UnstackableRepositoryFormat(self._format,
                                                 self.control_transport.base)

    def is_shared(self):
        return False

    def get_physical_lock_status(self):
        return False

    def lock_write(self):
        """See Branch.lock_write()."""
        if self._lock_mode:
            if self._lock_mode != 'w':
                raise errors.ReadOnlyError(self)
            self._lock_count += 1
        else:
            self._lock_mode = 'w'
            self._lock_count = 1
            self._transaction = transactions.WriteTransaction()
        return repository.RepositoryWriteLockResult(self.unlock, None)

    def break_lock(self):
        raise NotImplementedError(self.break_lock)

    def dont_leave_lock_in_place(self):
        raise NotImplementedError(self.dont_leave_lock_in_place)

    def leave_lock_in_place(self):
        raise NotImplementedError(self.leave_lock_in_place)

    def lock_read(self):
        if self._lock_mode:
            if self._lock_mode not in ('r', 'w'):
                raise AssertionError
            self._lock_count += 1
        else:
            self._lock_mode = 'r'
            self._lock_count = 1
            self._transaction = transactions.ReadOnlyTransaction()
        return lock.LogicalLockResult(self.unlock)

    @only_raises(errors.LockNotHeld, errors.LockBroken)
    def unlock(self):
        if self._lock_count == 0:
            raise errors.LockNotHeld(self)
        if self._lock_count == 1 and self._lock_mode == 'w':
            if self._write_group is not None:
                self.abort_write_group()
                self._lock_count -= 1
                self._lock_mode = None
                raise errors.BzrError(
                    'Must end write groups before releasing write locks.')
        self._lock_count -= 1
        if self._lock_count == 0:
            self._lock_mode = None
            transaction = self._transaction
            self._transaction = None
            transaction.finish()
            if hasattr(self, '_git'):
                self._git.close()

    def is_write_locked(self):
        return (self._lock_mode == 'w')

    def is_locked(self):
        return (self._lock_mode is not None)

    def get_transaction(self):
        """See Repository.get_transaction()."""
        if self._transaction is None:
            return transactions.PassThroughTransaction()
        else:
            return self._transaction

    def reconcile(self, other=None, thorough=False):
        """Reconcile this repository."""
        from ..reconcile import ReconcileResult
        ret = ReconcileResult()
        ret.aborted = False
        return ret

    def supports_rich_root(self):
        return True

    def get_mapping(self):
        return default_mapping

    def make_working_trees(self):
        raise NotImplementedError(self.make_working_trees)

    def revision_graph_can_have_wrong_parents(self):
        return False

    def add_signature_text(self, revid, signature):
        raise errors.UnsupportedOperation(self.add_signature_text, self)

    def sign_revision(self, revision_id, gpg_strategy):
        raise errors.UnsupportedOperation(self.add_signature_text, self)


class LocalGitRepository(GitRepository):
    """Git repository on the file system."""

    def __init__(self, gitdir):
        GitRepository.__init__(self, gitdir)
        self._git = gitdir._git
        self._file_change_scanner = GitFileLastChangeScanner(self)
        self._transaction = None

    def get_commit_builder(self, branch, parents, config, timestamp=None,
                           timezone=None, committer=None, revprops=None,
                           revision_id=None, lossy=False):
        """Obtain a CommitBuilder for this repository.

        :param branch: Branch to commit to.
        :param parents: Revision ids of the parents of the new revision.
        :param config: Configuration to use.
        :param timestamp: Optional timestamp recorded for commit.
        :param timezone: Optional timezone for timestamp.
        :param committer: Optional committer to set for commit.
        :param revprops: Optional dictionary of revision properties.
        :param revision_id: Optional revision id.
        :param lossy: Whether to discard data that can not be natively
            represented, when pushing to a foreign VCS
        """
        from .commit import GitCommitBuilder
        builder = GitCommitBuilder(
            self, parents, config, timestamp, timezone, committer, revprops,
            revision_id, lossy)
        self.start_write_group()
        return builder

    def _write_git_config(self, cs):
        f = BytesIO()
        cs.write_to_file(f)
        self._git._put_named_file('config', f.getvalue())

    def get_file_graph(self):
        return _mod_graph.Graph(GitFileParentProvider(
            self._file_change_scanner))

    def iter_files_bytes(self, desired_files):
        """Iterate through file versions.

        Files will not necessarily be returned in the order they occur in
        desired_files.  No specific order is guaranteed.

        Yields pairs of identifier, bytes_iterator.  identifier is an opaque
        value supplied by the caller as part of desired_files.  It should
        uniquely identify the file version in the caller's context.  (Examples:
        an index number or a TreeTransform trans_id.)

        bytes_iterator is an iterable of bytestrings for the file.  The
        kind of iterable and length of the bytestrings are unspecified, but for
        this implementation, it is a list of bytes produced by
        VersionedFile.get_record_stream().

        :param desired_files: a list of (file_id, revision_id, identifier)
            triples
        """
        per_revision = {}
        for (file_id, revision_id, identifier) in desired_files:
            per_revision.setdefault(revision_id, []).append(
                (file_id, identifier))
        for revid, files in per_revision.items():
            try:
                (commit_id, mapping) = self.lookup_bzr_revision_id(revid)
            except errors.NoSuchRevision:
                raise errors.RevisionNotPresent(revid, self)
            try:
                commit = self._git.object_store[commit_id]
            except KeyError:
                raise errors.RevisionNotPresent(revid, self)
            root_tree = commit.tree
            for fileid, identifier in files:
                try:
                    path = mapping.parse_file_id(fileid)
                except ValueError:
                    raise errors.RevisionNotPresent((fileid, revid), self)
                try:
                    mode, item_id = tree_lookup_path(
                        self._git.object_store.__getitem__, root_tree,
                        encode_git_path(path))
                    obj = self._git.object_store[item_id]
                except KeyError:
                    raise errors.RevisionNotPresent((fileid, revid), self)
                else:
                    if obj.type_name == b"tree":
                        yield (identifier, [])
                    elif obj.type_name == b"blob":
                        yield (identifier, obj.chunked)
                    else:
                        raise AssertionError("file text resolved to %r" % obj)

    def gather_stats(self, revid=None, committers=None):
        """See Repository.gather_stats()."""
        result = super().gather_stats(
            revid, committers)
        revs = []
        for sha in self._git.object_store:
            o = self._git.object_store[sha]
            if o.type_name == b"commit":
                revs.append(o.id)
        result['revisions'] = len(revs)
        return result

    def _iter_revision_ids(self):
        mapping = self.get_mapping()
        for sha in self._git.object_store:
            o = self._git.object_store[sha]
            if not isinstance(o, Commit):
                continue
            revid = mapping.revision_id_foreign_to_bzr(o.id)
            yield o.id, revid

    def all_revision_ids(self):
        ret = set()
        for git_sha, revid in self._iter_revision_ids():
            ret.add(revid)
        return list(ret)

    def _get_parents(self, revid, no_alternates=False):
        if not isinstance(revid, bytes):
            raise ValueError
        try:
            (hexsha, mapping) = self.lookup_bzr_revision_id(revid)
        except errors.NoSuchRevision:
            return None
        # FIXME: Honor no_alternates setting
        try:
            commit = self._git.object_store[hexsha]
        except KeyError:
            return None
        ret = []
        for p in commit.parents:
            try:
                ret.append(self.lookup_foreign_revision_id(p, mapping))
            except KeyError:
                ret.append(mapping.revision_id_foreign_to_bzr(p))
        return ret

    def _get_parent_map_no_fallbacks(self, revids):
        return self.get_parent_map(revids, no_alternates=True)

    def get_parent_map(self, revids, no_alternates=False):
        parent_map = {}
        for revision_id in revids:
            parents = self._get_parents(
                revision_id, no_alternates=no_alternates)
            if revision_id == _mod_revision.NULL_REVISION:
                parent_map[revision_id] = ()
                continue
            if parents is None:
                continue
            if len(parents) == 0:
                parents = [_mod_revision.NULL_REVISION]
            parent_map[revision_id] = tuple(parents)
        return parent_map

    def get_known_graph_ancestry(self, revision_ids):
        """Return the known graph for a set of revision ids and their ancestors.
        """
        pending = set(revision_ids)
        parent_map = {}
        while pending:
            this_parent_map = {}
            for revid in pending:
                if revid == _mod_revision.NULL_REVISION:
                    continue
                parents = self._get_parents(revid)
                if parents is not None:
                    this_parent_map[revid] = parents
            parent_map.update(this_parent_map)
            pending = set()
            for values in this_parent_map.values():
                pending.update(values)
            pending = pending.difference(parent_map)
        return _mod_graph.KnownGraph(parent_map)

    def get_signature_text(self, revision_id):
        git_commit_id, mapping = self.lookup_bzr_revision_id(revision_id)
        try:
            commit = self._git.object_store[git_commit_id]
        except KeyError:
            raise errors.NoSuchRevision(self, revision_id)
        if commit.gpgsig is None:
            raise errors.NoSuchRevision(self, revision_id)
        return commit.gpgsig

    def check(self, revision_ids=None, callback_refs=None, check_repo=True):
        result = GitCheck(self, check_repo=check_repo)
        result.check(callback_refs)
        return result

    def pack(self, hint=None, clean_obsolete_packs=False):
        self._git.object_store.pack_loose_objects()

    def lookup_foreign_revision_id(self, foreign_revid, mapping=None):
        """Lookup a revision id.

        :param foreign_revid: Foreign revision id to look up
        :param mapping: Mapping to use (use default mapping if not specified)
        :raise KeyError: If foreign revision was not found
        :return: bzr revision id
        """
        if not isinstance(foreign_revid, bytes):
            raise TypeError(foreign_revid)
        if mapping is None:
            mapping = self.get_mapping()
        if foreign_revid == ZERO_SHA:
            return _mod_revision.NULL_REVISION
        unpeeled, peeled = peel_sha(self._git.object_store, foreign_revid)
        if not isinstance(peeled, Commit):
            raise NotCommitError(peeled.id)
        revid = mapping.get_revision_id(peeled)
        # FIXME: check testament before doing this?
        return revid

    def has_signature_for_revision_id(self, revision_id):
        """Check whether a GPG signature is present for this revision.

        This is never the case for Git repositories.
        """
        try:
            self.get_signature_text(revision_id)
        except errors.NoSuchRevision:
            return False
        else:
            return True

    def verify_revision_signature(self, revision_id, gpg_strategy):
        """Verify the signature on a revision.

        :param revision_id: the revision to verify
        :gpg_strategy: the GPGStrategy object to used

        :return: gpg.SIGNATURE_VALID or a failed SIGNATURE_ value
        """
        from breezy import gpg
        with self.lock_read():
            git_commit_id, mapping = self.lookup_bzr_revision_id(revision_id)
            try:
                commit = self._git.object_store[git_commit_id]
            except KeyError:
                raise errors.NoSuchRevision(self, revision_id)

            if commit.gpgsig is None:
                return gpg.SIGNATURE_NOT_SIGNED, None

            without_sig = Commit.from_string(commit.as_raw_string())
            without_sig.gpgsig = None

            (result, key, plain_text) = gpg_strategy.verify(
                without_sig.as_raw_string(), commit.gpgsig)
            return (result, key)

    def lookup_bzr_revision_id(self, bzr_revid, mapping=None):
        """Lookup a bzr revision id in a Git repository.

        :param bzr_revid: Bazaar revision id
        :param mapping: Optional mapping to use
        :return: Tuple with git commit id, mapping that was used and supplement
            details
        """
        try:
            (git_sha, mapping) = mapping_registry.revision_id_bzr_to_foreign(
                bzr_revid)
        except errors.InvalidRevisionId:
            raise errors.NoSuchRevision(self, bzr_revid)
        else:
            return (git_sha, mapping)

    def get_revision(self, revision_id):
        if not isinstance(revision_id, bytes):
            raise errors.InvalidRevisionId(revision_id, self)
        git_commit_id, mapping = self.lookup_bzr_revision_id(revision_id)
        try:
            commit = self._git.object_store[git_commit_id]
        except KeyError:
            raise errors.NoSuchRevision(self, revision_id)
        revision, roundtrip_revid, verifiers = mapping.import_commit(
            commit, self.lookup_foreign_revision_id, strict=False)
        if revision is None:
            raise AssertionError
        # FIXME: check verifiers ?
        if roundtrip_revid:
            revision.revision_id = roundtrip_revid
        return revision

    def has_revision(self, revision_id):
        """See Repository.has_revision."""
        if revision_id == _mod_revision.NULL_REVISION:
            return True
        try:
            git_commit_id, mapping = self.lookup_bzr_revision_id(revision_id)
        except errors.NoSuchRevision:
            return False
        return (git_commit_id in self._git)

    def has_revisions(self, revision_ids):
        """See Repository.has_revisions."""
        return set(filter(self.has_revision, revision_ids))

    def iter_revisions(self, revision_ids):
        """See Repository.get_revisions."""
        for revid in revision_ids:
            try:
                rev = self.get_revision(revid)
            except errors.NoSuchRevision:
                rev = None
            yield (revid, rev)

    def revision_trees(self, revids):
        """See Repository.revision_trees."""
        for revid in revids:
            yield self.revision_tree(revid)

    def revision_tree(self, revision_id):
        """See Repository.revision_tree."""
        if revision_id is None:
            raise ValueError('invalid revision id %s' % revision_id)
        return GitRevisionTree(self, revision_id)

    def set_make_working_trees(self, trees):
        raise errors.UnsupportedOperation(self.set_make_working_trees, self)

    def make_working_trees(self):
        return not self._git.get_config().get_boolean(("core", ), "bare")


class GitRepositoryFormat(repository.RepositoryFormat):
    """Git repository format."""

    supports_versioned_directories = False
    supports_tree_reference = True
    rich_root_data = True
    supports_leaving_lock = False
    fast_deltas = True
    supports_funky_characters = True
    supports_external_lookups = False
    supports_full_versioned_files = False
    supports_revision_signatures = False
    supports_nesting_repositories = False
    revision_graph_can_have_wrong_parents = False
    supports_unreferenced_revisions = True
    supports_setting_revision_ids = False
    supports_storing_branch_nick = False
    supports_overriding_transport = False
    supports_custom_revision_properties = False
    records_per_file_revision = False
    supports_multiple_authors = False
    supports_ghosts = False
    supports_chks = False

    @property
    def _matchingcontroldir(self):
        from .dir import LocalGitControlDirFormat
        return LocalGitControlDirFormat()

    def get_format_description(self):
        return "Git Repository"

    def initialize(self, controldir, shared=False, _internal=False):
        from .dir import GitDir
        if not isinstance(controldir, GitDir):
            raise errors.UninitializableFormat(self)
        return controldir.open_repository()

    def check_conversion_target(self, target_repo_format):
        return target_repo_format.rich_root_data

    def get_foreign_tests_repository_factory(self):
        from .tests.test_repository import ForeignTestsRepositoryFactory
        return ForeignTestsRepositoryFactory()

    def network_name(self):
        return b"git"


def get_extra_interrepo_test_combinations():
    from ..bzr.groupcompress_repo import RepositoryFormat2a
    from . import interrepo
    return [
        (interrepo.InterLocalGitNonGitRepository,
         GitRepositoryFormat(), RepositoryFormat2a()),
        (interrepo.InterLocalGitLocalGitRepository,
         GitRepositoryFormat(), GitRepositoryFormat()),
        (interrepo.InterToLocalGitRepository,
         RepositoryFormat2a(), GitRepositoryFormat()),
        ]



# Copyright (C) 2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for OS Locks."""


from .. import debug, errors, lock, tests
from .scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class TestOSLock(tests.TestCaseInTempDir):

    scenarios = [(
        name, {
            'write_lock': write_lock,
            'read_lock': read_lock})
        for name, write_lock, read_lock in lock._lock_classes]

    read_lock = None
    write_lock = None

    def setUp(self):
        super().setUp()
        self.build_tree(['a-lock-file'])

    def test_create_read_lock(self):
        r_lock = self.read_lock('a-lock-file')
        r_lock.unlock()

    def test_create_write_lock(self):
        w_lock = self.write_lock('a-lock-file')
        w_lock.unlock()

    def test_read_locks_share(self):
        r_lock = self.read_lock('a-lock-file')
        try:
            lock2 = self.read_lock('a-lock-file')
            lock2.unlock()
        finally:
            r_lock.unlock()

    def test_write_locks_are_exclusive(self):
        w_lock = self.write_lock('a-lock-file')
        try:
            self.assertRaises(errors.LockContention,
                              self.write_lock, 'a-lock-file')
        finally:
            w_lock.unlock()

    def test_read_locks_block_write_locks(self):
        r_lock = self.read_lock('a-lock-file')
        try:
            if lock.have_fcntl and self.write_lock is lock._fcntl_WriteLock:
                # With -Dlock, fcntl locks are properly exclusive
                debug.debug_flags.add('strict_locks')
                self.assertRaises(errors.LockContention,
                                  self.write_lock, 'a-lock-file')
                # But not without it
                debug.debug_flags.remove('strict_locks')
                try:
                    w_lock = self.write_lock('a-lock-file')
                except errors.LockContention:
                    self.fail('Unexpected success. fcntl read locks'
                              ' do not usually block write locks')
                else:
                    w_lock.unlock()
                    self.knownFailure('fcntl read locks don\'t'
                                      ' block write locks without -Dlock')
            else:
                self.assertRaises(errors.LockContention,
                                  self.write_lock, 'a-lock-file')
        finally:
            r_lock.unlock()

    def test_write_locks_block_read_lock(self):
        w_lock = self.write_lock('a-lock-file')
        try:
            if lock.have_fcntl and self.read_lock is lock._fcntl_ReadLock:
                # With -Dlock, fcntl locks are properly exclusive
                debug.debug_flags.add('strict_locks')
                self.assertRaises(errors.LockContention,
                                  self.read_lock, 'a-lock-file')
                # But not without it
                debug.debug_flags.remove('strict_locks')
                try:
                    r_lock = self.read_lock('a-lock-file')
                except errors.LockContention:
                    self.fail('Unexpected success. fcntl write locks'
                              ' do not usually block read locks')
                else:
                    r_lock.unlock()
                    self.knownFailure('fcntl write locks don\'t'
                                      ' block read locks without -Dlock')
            else:
                self.assertRaises(errors.LockContention,
                                  self.read_lock, 'a-lock-file')
        finally:
            w_lock.unlock()

    def test_temporary_write_lock(self):
        r_lock = self.read_lock('a-lock-file')
        try:
            status, w_lock = r_lock.temporary_write_lock()
            self.assertTrue(status)
            # This should block another write lock
            try:
                self.assertRaises(errors.LockContention,
                                  self.write_lock, 'a-lock-file')
            finally:
                r_lock = w_lock.restore_read_lock()
            # We should be able to take a read lock now
            r_lock2 = self.read_lock('a-lock-file')
            r_lock2.unlock()
        finally:
            r_lock.unlock()

    def test_temporary_write_lock_fails(self):
        r_lock = self.read_lock('a-lock-file')
        try:
            r_lock2 = self.read_lock('a-lock-file')
            try:
                status, w_lock = r_lock.temporary_write_lock()
                self.assertFalse(status)
                # Taking out the lock requires unlocking and locking again, so
                # we have to replace the original object
                r_lock = w_lock
            finally:
                r_lock2.unlock()
            # We should be able to take a read lock now
            r_lock2 = self.read_lock('a-lock-file')
            r_lock2.unlock()
        finally:
            r_lock.unlock()



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for ContentFilterTree"""

import tarfile
import zipfile

from .. import export, filter_tree, tests
from . import fixtures
from .test_filters import _stack_1


class TestFilterTree(tests.TestCaseWithTransport):

    def make_tree(self):
        self.underlying_tree = fixtures.make_branch_and_populated_tree(
            self)

        def stack_callback(path):
            return _stack_1
        self.filter_tree = filter_tree.ContentFilterTree(
            self.underlying_tree, stack_callback)
        return self.filter_tree

    def test_get_file_text(self):
        self.make_tree()
        self.assertEqual(
            self.underlying_tree.get_file_text('hello'),
            b'hello world')
        self.assertEqual(
            self.filter_tree.get_file_text('hello'),
            b'HELLO WORLD')

    def test_tar_export_content_filter_tree(self):
        # TODO: this could usefully be run generically across all exporters.
        self.make_tree()
        export.export(self.filter_tree, "out.tgz")
        ball = tarfile.open("out.tgz", "r:gz")
        self.assertEqual(
            b'HELLO WORLD',
            ball.extractfile('out/hello').read())

    def test_zip_export_content_filter_tree(self):
        self.make_tree()
        export.export(self.filter_tree, 'out.zip')
        zipf = zipfile.ZipFile('out.zip', 'r')
        self.assertEqual(
            b'HELLO WORLD',
            zipf.read('out/hello'))



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A convenience class around smtplib."""

import errno
import smtplib
import socket
from email.utils import getaddresses, parseaddr

from . import config, osutils
from .errors import BzrError, InternalBzrError

smtp_password = config.Option('smtp_password', default=None,
                              help='''\
Password to use for authentication to SMTP server.
''')
smtp_server = config.Option('smtp_server', default=None,
                            help='''\
Hostname of the SMTP server to use for sending email.
''')
smtp_username = config.Option('smtp_username', default=None,
                              help='''\
Username to use for authentication to SMTP server.
''')


class SMTPError(BzrError):

    _fmt = "SMTP error: %(error)s"

    def __init__(self, error):
        self.error = error


class SMTPConnectionRefused(SMTPError):

    _fmt = "SMTP connection to %(host)s refused"

    def __init__(self, error, host):
        self.error = error
        self.host = host


class DefaultSMTPConnectionRefused(SMTPConnectionRefused):

    _fmt = "Please specify smtp_server.  No server at default %(host)s."


class NoDestinationAddress(InternalBzrError):

    _fmt = "Message does not have a destination address."


class SMTPConnection:
    """Connect to an SMTP server and send an email.

    This is a gateway between breezy.config.Config and smtplib.SMTP. It
    understands the basic bzr SMTP configuration information: smtp_server,
    smtp_username, and smtp_password.
    """

    _default_smtp_server = 'localhost'

    def __init__(self, config, _smtp_factory=None):
        self._smtp_factory = _smtp_factory
        if self._smtp_factory is None:
            self._smtp_factory = smtplib.SMTP
        self._config = config
        self._config_smtp_server = config.get('smtp_server')
        self._smtp_server = self._config_smtp_server
        if self._smtp_server is None:
            self._smtp_server = self._default_smtp_server

        self._smtp_username = config.get('smtp_username')
        self._smtp_password = config.get('smtp_password')

        self._connection = None

    def _connect(self):
        """If we haven't connected, connect and authenticate."""
        if self._connection is not None:
            return

        self._create_connection()
        # FIXME: _authenticate() should only be called when the server has
        # refused unauthenticated access, so it can safely try to authenticate
        # with the default username. JRV20090407
        self._authenticate()

    def _create_connection(self):
        """Create an SMTP connection."""
        try:
            self._connection = self._smtp_factory(host=self._smtp_server)
        except OSError as e:
            if e.args[0] == errno.ECONNREFUSED:
                if self._config_smtp_server is None:
                    raise DefaultSMTPConnectionRefused(socket.error,
                                                       self._smtp_server)
                else:
                    raise SMTPConnectionRefused(socket.error,
                                                self._smtp_server)
            else:
                raise

        # Say EHLO (falling back to HELO) to query the server's features.
        code, resp = self._connection.ehlo()
        if not (200 <= code <= 299):
            code, resp = self._connection.helo()
            if not (200 <= code <= 299):
                raise SMTPError("server refused HELO: %d %s" % (code, resp))

        # Use TLS if the server advertised it:
        if self._connection.has_extn("starttls"):
            code, resp = self._connection.starttls()
            if not (200 <= code <= 299):
                raise SMTPError("server refused STARTTLS: %d %s" %
                                (code, resp))
            # Say EHLO again, to check for newly revealed features
            code, resp = self._connection.ehlo()
            if not (200 <= code <= 299):
                raise SMTPError("server refused EHLO: %d %s" % (code, resp))

    def _authenticate(self):
        """If necessary authenticate yourself to the server."""
        auth = config.AuthenticationConfig()
        if self._smtp_username is None:
            # FIXME: Since _authenticate gets called even when no authentication
            # is necessary, it's not possible to use the default username
            # here yet.
            self._smtp_username = auth.get_user('smtp', self._smtp_server)
            if self._smtp_username is None:
                return

        if self._smtp_password is None:
            self._smtp_password = auth.get_password(
                'smtp', self._smtp_server, self._smtp_username)

        # smtplib requires that the username and password be byte
        # strings.  The CRAM-MD5 spec doesn't give any guidance on
        # encodings, but the SASL PLAIN spec says UTF-8, so that's
        # what we'll use.
        username = osutils.safe_utf8(self._smtp_username)
        password = osutils.safe_utf8(self._smtp_password)

        self._connection.login(username, password)

    @staticmethod
    def get_message_addresses(message):
        """Get the origin and destination addresses of a message.

        :param message: A message object supporting get() to access its
            headers, like email.message.Message or
            breezy.email_message.EmailMessage.
        :return: A pair (from_email, to_emails), where from_email is the email
            address in the From header, and to_emails a list of all the
            addresses in the To, Cc, and Bcc headers.
        """
        from_email = parseaddr(message.get('From', None))[1]
        to_full_addresses = []
        for header in ['To', 'Cc', 'Bcc']:
            value = message.get(header, None)
            if value:
                to_full_addresses.append(value)
        to_emails = [pair[1] for pair in
                     getaddresses(to_full_addresses)]

        return from_email, to_emails

    def send_email(self, message):
        """Send an email message.

        The message will be sent to all addresses in the To, Cc and Bcc
        headers.

        :param message: An email.message.Message or
            email.mime.multipart.MIMEMultipart object.
        :return: None
        """
        from_email, to_emails = self.get_message_addresses(message)

        if not to_emails:
            raise NoDestinationAddress

        try:
            self._connect()
            self._connection.sendmail(from_email, to_emails,
                                      message.as_string())
        except smtplib.SMTPRecipientsRefused as e:
            raise SMTPError('server refused recipient: %d %s' %
                            next(iter(e.recipients.values())))
        except smtplib.SMTPResponseException as e:
            raise SMTPError('%d %s' % (e.smtp_code, e.smtp_error))
        except smtplib.SMTPException as e:
            raise SMTPError(str(e))



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test branches with inaccessible parents."""

from breezy import branch, errors
from breezy.tests import per_branch, test_server


class InaccessibleParentTests(per_branch.TestCaseWithBranch):
    """Tests with branches with "inaccessible" parents.

    An "inaccessible" parent location is one that cannot be represented, e.g. if
    a child branch says its parent is at "../../my-parent", but that child is at
    "http://host/one" then that parent location is inaccessible.  These
    branches' get_parent method will raise InaccessibleParent.
    """

    def setUp(self):
        super().setUp()
        if self.transport_server in (test_server.LocalURLServer, None):
            self.transport_readonly_server = test_server.TestingChrootServer

    def get_branch_with_invalid_parent(self):
        """Get a branch whose get_parent will raise InaccessibleParent."""
        self.build_tree(
            ['parent/', 'parent/path/', 'parent/path/to/',
             'child/', 'child/path/', 'child/path/to/'],
            transport=self.get_transport())
        self.make_branch(
            'parent/path/to/a').controldir.sprout(self.get_url('child/path/to/b'))

        # The child branch internally will have recorded that its parent is at
        # "../../../../parent/path/to/a" or similar.  So we move the child
        # branch up several directories, so that its parent path will point to
        # somewhere outside the directory served by the HTTP server.  Thus its
        # parent is now inaccessible.
        self.get_transport().rename('child/path/to/b', 'b')
        branch_b = branch.Branch.open(self.get_readonly_url('b'))
        return branch_b

    def test_get_parent_invalid(self):
        # When you have a branch whose parent URL cannot be calculated, this
        # exception will be raised.
        branch_b = self.get_branch_with_invalid_parent()
        self.assertRaises(errors.InaccessibleParent, branch_b.get_parent)

    def test_clone_invalid_parent(self):
        # If clone can't determine the location of the parent of the branch
        # being cloned, then the new branch will have no parent set.
        branch_b = self.get_branch_with_invalid_parent()
        branch_c = branch_b.controldir.clone('c').open_branch()
        self.assertEqual(None, branch_c.get_parent())

    def test_sprout_invalid_parent(self):
        # A sprouted branch will have a parent of the branch it was sprouted
        # from, even if that branch has an invalid parent.
        branch_b = self.get_branch_with_invalid_parent()
        branch_c = branch_b.controldir.sprout('c').open_branch()
        self.assertEqual(branch_b.base, branch_c.get_parent())



# Copyright (C) 2019 Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License or
# (at your option) a later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Subversion foreign branch support.

Currently only tells the user that Subversion is not supported.
"""

from ... import version_info  # noqa: F401
from ... import controldir, errors
from ... import transport as _mod_transport
from ...revisionspec import revspec_registry


class SubversionUnsupportedError(errors.UnsupportedVcs):

    vcs = "svn"

    _fmt = ('Subversion branches are not yet supported. '
            'To interoperate with Subversion branches, use fastimport.')


class SvnWorkingTreeDirFormat(controldir.ControlDirFormat):
    """Subversion directory format."""

    def get_converter(self):
        raise NotImplementedError(self.get_converter)

    def get_format_description(self):
        return "Subversion working directory"

    def initialize_on_transport(self, transport):
        raise errors.UninitializableFormat(format=self)

    def is_supported(self):
        return False

    def supports_transport(self, transport):
        return False

    def check_support_status(self, allow_unsupported, recommend_upgrade=True,
                             basedir=None):
        raise SubversionUnsupportedError(format=self)

    def open(self, transport):
        # Raise NotBranchError if there is nothing there
        SvnWorkingTreeProber().probe_transport(transport)
        raise NotImplementedError(self.open)


class SvnWorkingTreeProber(controldir.Prober):

    @classmethod
    def priority(klass, transport):
        return 100

    def probe_transport(self, transport):
        try:
            transport.local_abspath('.')
        except errors.NotLocalUrl:
            raise errors.NotBranchError(path=transport.base)
        else:
            if transport.has(".svn"):
                return SvnWorkingTreeDirFormat()
            raise errors.NotBranchError(path=transport.base)

    @classmethod
    def known_formats(cls):
        return [SvnWorkingTreeDirFormat()]


class SvnRepositoryFormat(controldir.ControlDirFormat):
    """Subversion directory format."""

    def get_converter(self):
        raise NotImplementedError(self.get_converter)

    def get_format_description(self):
        return "Subversion repository"

    def initialize_on_transport(self, transport):
        raise errors.UninitializableFormat(self)

    def is_supported(self):
        return False

    def supports_transport(self, transport):
        return False

    def check_support_status(self, allow_unsupported, recommend_upgrade=True,
                             basedir=None):
        raise SubversionUnsupportedError()

    def open(self, transport):
        # Raise NotBranchError if there is nothing there
        SvnRepositoryProber().probe_transport(transport)
        raise NotImplementedError(self.open)


class SvnRepositoryProber(controldir.Prober):

    _supported_schemes = ["http", "https", "file", "svn"]

    @classmethod
    def priority(klass, transport):
        if 'svn' in transport.base:
            return 90
        return 100

    def probe_transport(self, transport):
        try:
            url = transport.external_url()
        except errors.InProcessTransport:
            raise errors.NotBranchError(path=transport.base)

        scheme = url.split(":")[0]
        if scheme.startswith("svn+") or scheme == "svn":
            raise SubversionUnsupportedError()

        if scheme not in self._supported_schemes:
            raise errors.NotBranchError(path=transport.base)

        if scheme == 'file':
            # Cheaper way to figure out if there is a svn repo
            maybe = False
            subtransport = transport
            while subtransport:
                try:
                    if all([subtransport.has(name)
                            for name in ["format", "db", "conf"]]):
                        maybe = True
                        break
                except UnicodeEncodeError:
                    pass
                prevsubtransport = subtransport
                subtransport = prevsubtransport.clone("..")
                if subtransport.base == prevsubtransport.base:
                    break
            if not maybe:
                raise errors.NotBranchError(path=transport.base)

        # If this is a HTTP transport, use the existing connection to check
        # that the remote end supports version control.
        if scheme in ("http", "https"):
            priv_transport = getattr(transport, "_decorated", transport)
            try:
                headers = priv_transport._options('.')
            except (errors.InProcessTransport, _mod_transport.NoSuchFile,
                    errors.InvalidHttpResponse):
                raise errors.NotBranchError(path=transport.base)
            else:
                dav_entries = set()
                for key, value in headers:
                    if key.upper() == 'DAV':
                        dav_entries.update(
                            [x.strip() for x in value.split(',')])
                if "version-control" not in dav_entries:
                    raise errors.NotBranchError(path=transport.base)

        return SvnRepositoryFormat()

    @classmethod
    def known_formats(cls):
        return [SvnRepositoryFormat()]


controldir.ControlDirFormat.register_prober(SvnWorkingTreeProber)
controldir.ControlDirFormat.register_prober(SvnRepositoryProber)


revspec_registry.register_lazy("svn:", __name__ + ".revspec", "RevisionSpec_svn")


_mod_transport.register_transport_proto(
    'svn+ssh://',
    help="Access using the Subversion smart server tunneled over SSH.")
_mod_transport.register_transport_proto(
    'svn+http://')
_mod_transport.register_transport_proto(
    'svn+https://')
_mod_transport.register_transport_proto(
    'svn://',
    help="Access using the Subversion smart server.")



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Branch.revision_id_to_dotted_revno()"""

from breezy import errors
from breezy.tests.per_branch import TestCaseWithBranch


class TestRevisionIdToDottedRevno(TestCaseWithBranch):

    def test_lookup_dotted_revno(self):
        tree, revmap = self.create_tree_with_merge()
        the_branch = tree.branch
        self.assertEqual(
            (0,), the_branch.revision_id_to_dotted_revno(b'null:'))
        self.assertEqual(
            (1,), the_branch.revision_id_to_dotted_revno(revmap['1']))
        self.assertEqual(
            (2,), the_branch.revision_id_to_dotted_revno(revmap['2']))
        self.assertEqual(
            (3,), the_branch.revision_id_to_dotted_revno(revmap['3']))
        self.assertEqual((1, 1, 1), the_branch.revision_id_to_dotted_revno(
            revmap['1.1.1']))
        self.assertRaises(errors.NoSuchRevision,
                          the_branch.revision_id_to_dotted_revno, b'rev-1.0.2')



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Base implementation of Transport over http.

There are separate implementation modules for each http client implementation.
"""

DEBUG = 0

import os
import ssl
import sys

from ... import config
from ... import version_string as breezy_version


def default_user_agent():
    return 'Breezy/%s' % breezy_version


# Note for packagers: if there is no package providing certs for your platform,
# the curl project produces http://curl.haxx.se/ca/cacert.pem weekly.
_ssl_ca_certs_known_locations = [
    '/etc/ssl/certs/ca-certificates.crt',  # Ubuntu/debian/gentoo
    '/etc/pki/tls/certs/ca-bundle.crt',  # Fedora/CentOS/RH
    '/etc/ssl/ca-bundle.pem',  # OpenSuse
    '/etc/ssl/cert.pem',  # OpenSuse
    "/usr/local/share/certs/ca-root-nss.crt",  # FreeBSD
    # XXX: Needs checking, can't trust the interweb ;) -- vila 2012-01-25
    '/etc/openssl/certs/ca-certificates.crt',  # Solaris
]


def default_ca_certs():
    if sys.platform == 'win32':
        return os.path.join(os.path.dirname(sys.executable), "cacert.pem")
    elif sys.platform == 'darwin':
        # FIXME: Needs some default value for osx, waiting for osx installers
        # guys feedback -- vila 2012-01-25
        pass
    else:
        # Try known locations for friendly OSes providing the root certificates
        # without making them hard to use for any https client.
        for path in _ssl_ca_certs_known_locations:
            if os.path.exists(path):
                # First found wins
                return path
    # A default path that makes sense and will be mentioned in the error
    # presented to the user, even if not correct for all platforms
    return _ssl_ca_certs_known_locations[0]


def ca_certs_from_store(path):
    if not os.path.exists(path):
        raise ValueError("ca certs path %s does not exist" % path)
    return path


def cert_reqs_from_store(unicode_str):
    import ssl
    try:
        return {"required": ssl.CERT_REQUIRED,
                "none": ssl.CERT_NONE}[unicode_str]
    except KeyError:
        raise ValueError("invalid value %s" % unicode_str)


def default_ca_reqs():
    if sys.platform in ('win32', 'darwin'):
        # FIXME: Once we get a native access to root certificates there, this
        # won't needed anymore. See http://pad.lv/920455 -- vila 2012-02-15
        return 'none'
    else:
        return 'required'


opt_ssl_ca_certs = config.Option('ssl.ca_certs',
                                 from_unicode=ca_certs_from_store,
                                 default=default_ca_certs,
                                 invalid='warning',
                                 help="""\
Path to certification authority certificates to trust.

This should be a valid path to a bundle containing all root Certificate
Authorities used to verify an https server certificate.

Use ssl.cert_reqs=none to disable certificate verification.
""")

opt_ssl_cert_reqs = config.Option('ssl.cert_reqs',
                                  default=default_ca_reqs,
                                  from_unicode=cert_reqs_from_store,
                                  invalid='error',
                                  help="""\
Whether to require a certificate from the remote side. (default:required)

Possible values:
 * none: Certificates ignored
 * required: Certificates required and validated
""")



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for implementations of Repository.has_same_location."""

from breezy import controldir, transport
from breezy.tests import TestNotApplicable
from breezy.tests.per_repository import TestCaseWithRepository


class TestHasSameLocation(TestCaseWithRepository):
    """Tests for Repository.has_same_location method."""

    def assertSameRepo(self, a, b):
        """Asserts that two objects are the same repository.

        This method does the comparison both ways (`a.has_same_location(b)` as
        well as `b.has_same_location(a)`) to make sure both objects'
        `has_same_location` methods give the same results.
        """
        self.assertTrue(a.has_same_location(b),
                        "{!r} is not the same repository as {!r}".format(a, b))
        self.assertTrue(b.has_same_location(a),
                        "{!r} is the same as {!r}, but not vice versa".format(a, b))

    def assertDifferentRepo(self, a, b):
        """Asserts that two objects are the not same repository.

        This method does the comparison both ways (`a.has_same_location(b)` as
        well as `b.has_same_location(a)`) to make sure both objects'
        `has_same_location` methods give the same results.

        :seealso: assertDifferentRepo
        """
        self.assertFalse(a.has_same_location(b),
                         "{!r} is not the same repository as {!r}".format(a, b))
        self.assertFalse(b.has_same_location(a),
                         "{!r} is the same as {!r}, but not vice versa".format(a, b))

    def test_same_repo_instance(self):
        """A repository object is the same repository as itself."""
        repo = self.make_repository('.')
        self.assertSameRepo(repo, repo)

    def test_same_repo_location(self):
        """Different repository objects for the same location are the same."""
        repo = self.make_repository('.')
        reopened_repo = repo.controldir.open_repository()
        self.assertFalse(
            repo is reopened_repo,
            "This test depends on reopened_repo being a different instance of "
            "the same repo.")
        self.assertSameRepo(repo, reopened_repo)

    def test_different_repos_not_equal(self):
        """Repositories at different locations are not the same."""
        repo_one = self.make_repository('one')
        repo_two = self.make_repository('two')
        self.assertDifferentRepo(repo_one, repo_two)

    def test_same_bzrdir_different_control_files_not_equal(self):
        """Repositories in the same bzrdir, but with different control files,
        are not the same.

        This can happens e.g. when upgrading a repository.  This test mimics how
        CopyConverter creates a second repository in one bzrdir.
        """
        repo = self.make_repository('repo')
        repo.control_transport.copy_tree('.', '../repository.backup')
        backup_transport = repo.control_transport.clone('../repository.backup')
        if not repo._format.supports_overriding_transport:
            raise TestNotApplicable("remote repositories don't support overriding "
                                    "transport")
        backup_repo = repo._format.open(repo.controldir,
                                        _override_transport=backup_transport)
        self.assertDifferentRepo(repo, backup_repo)

    def test_different_format_not_equal(self):
        """Different format repositories are comparable and not the same.

        Comparing different format repository objects should give a negative
        result, rather than trigger an exception (which could happen with a
        naive __eq__ implementation, e.g. due to missing attributes).
        """
        repo = self.make_repository('repo')
        other_repo = self.make_repository('other', format='default')
        if repo._format == other_repo._format:
            # We're testing the default format!  So we have to use a non-default
            # format for other_repo.
            transport.get_transport_from_url(
                self.get_vfs_only_url()).delete_tree('other')
            other_repo = self.make_repository('other', format='knit')
        # Make sure the other_repo is not a RemoteRepository.
        other_bzrdir = controldir.ControlDir.open(
            self.get_vfs_only_url('other'))
        other_repo = other_bzrdir.open_repository()
        self.assertDifferentRepo(repo, other_repo)
